__all__ = ('AbstractBidirectionalTransportLayerBase', 'AbstractProtocolBase', 'AbstractTransportLayerBase', )

class AbstractTransportLayerBase:
    """
    Defines abstract transport layer functionality.
    """
    __slots__ = ()
    
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any`, `None`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return default
    
    
    def get_protocol(self):
        """
        Gets the transport's actual protocol.
        
        Returns
        -------
        protocol : `None`, ``ProtocolBase``
            Asynchronous protocol implementation.
        """
        return None
    
    
    def set_protocol(self, protocol):
        """
        Sets a new protocol to the transport.
        
        Parameters
        ----------
        protocol : ``ProtocolBase``
            Asynchronous protocol implementation.
        """
        pass
    
    
    def is_closing(self):
        """
        Returns whether the transport is closing.
        
        Returns
        -------
        is_closing : `bool`
        """
        return True

    
    def close(self):
        """
        Starts the shutdown process of the transport.
        
        If the transport is already closing, does nothing.
        """
        pass
    
    
    def abort(self):
        """
        Closes the transport immediately.
        
        The buffered data will be lost.
        """
        pass
    
    
    def write(self, data):
        """
        Write the given `data` to the transport.
        
        Do not blocks, but queues up the data instead to be sent as can.
        
        Parameters
        ----------
        data : `bytes-like`
            The data to send.
        
        Raises
        ------
        TypeError
            If `data` was not given as `bytes-like`.
        RuntimeError : `bool`
            If ``.write_eof`` was already called.
        """
        pass
    
    
    def writelines(self, lines):
        """
        Writes the given lines to the transport's socket.
        
        Parameters
        ----------
        lines : `iterable` of `bytes-like`
            The lines to write.
        
        Raises
        ------
        RuntimeError
            Protocol has no attached transport.
        """
        self.write(b''.join(lines))
    
    
    def write_eof(self):
        """
        Writes eof to the subprocess pipe's transport's protocol if applicable.
        
        By default ``SubprocessStreamWriter``'s transport is ``UnixWritePipeTransport``, what will call connection lost
        as well when the write buffer is empty.
        """
        pass
    
    
    def can_write_eof(self):
        """
        Return whether the transport supports ``.write_eof``.
        
        Returns
        -------
        can_write_eof : `bool`
        """
        return False
    
    
    def get_write_buffer_size(self):
        """
        Return the current size of the write buffer.
        
        Returns
        -------
        get_write_buffer_size : `int`
        """
        return 0
    
    
    def get_write_buffer_limits(self):
        """
        Returns the low and the high water of the transport.
        
        Returns
        -------
        low_water : `int`
            The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.
        high_water : `int`
            The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to
            `65536`.
        """
        return 0, 0
    
    
    def set_write_buffer_limits(self, high=None, low=None):
        """
        Set the high- and low-water limits for write flow control.
        
        These two values control when to call the protocol's ``.pause_writing`` and ``.resume_writing`` methods. If
        specified, the low-water limit must be less than or equal to the high-water limit. Neither value can be
        negative. The defaults are implementation-specific. If only the high-water limit is given, the low-water limit
        defaults to an implementation-specific value less than or equal to the high-water limit. Setting high to zero
        forces low to zero as well, and causes ``.pause_writing`` to be called whenever the buffer becomes non-empty.
        Setting low to zero causes ``.resume_writing`` to be called only once the buffer is empty. Use of zero for
        either limit is generally sub-optimal as it reduces opportunities for doing I/O and computation concurrently.
        
        Parameters
        ----------
        high : `None`, `int` = `None`, Optional
            High limit to stop reading if reached.
        low : `None`, `int` = `None`, Optional
            Low limit to start reading if reached.
        
        Raises
        ------
        ValueError
            If `low` is lower than `0` or if `low` is higher than `high`.
        """
        pass
    
    
    def pause_reading(self):
        """
        Pauses the receiving end.
        
        No data will be passed to the respective protocol's ``.data_received`` method until ``.resume_reading`` is
        called.
        
        Returns
        -------
        reading_paused : `bool`
            Whether reading was paused.
        """
        return False
    
    
    def resume_reading(self):
        """
        Resumes the receiving end.
        
        Data received will once again be passed to the respective protocol's ``.data_received`` method.
        
        Returns
        -------
        reading_resumed : `bool`
            Whether reading was resumed.
        """
        return False


class AbstractProtocolBase:
    """
    Defines abstract protocol functionality.
    
    Structure
    ---------
    All protocol should implement the following methods:
    - ``.connection_made``
    - ``.connection_lost``
    - ``.close``
    - ``.close_transport``
    - ``.get_transport``
    - ``.get_extra_info``
    
    Read protocols should implement the following:
    - ``.set_exception``
    - ``.eof_received``
    - ``.data_received``
    
    Write protocols should implement the following:
    - ``.pause_writing``
    - ``.resume_writing``
    - ``.write``
    - ``.writelines``
    - ``.write_eof``
    - ``.can_write_eof``
    - ``.drain``
    
    Datagram protocols should implement the following:
    - ``.datagram_received``
    - ``.error_received``
    """
    __slots__ = ()
    
    def connection_made(self, transport):
        """
        Called when a connection is made.
        
        Parameters
        ----------
        transport : `Any`
            Asynchronous transport implementation, what calls the protocol's ``.data_received`` when data is
            received.
        """
        pass
    
    
    def connection_lost(self, exception):
        """
        Called when the connection is lost or closed.
        
        Parameters
        ----------
        exception : `None`, `BaseException`
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        pass
    
    
    def close(self):
        """
        Closes the protocol by closing it's transport if applicable.
        """
        pass
    
    
    def close_transport(self, force=False):
        """
        Starts the shutdown process of the protocol's transport if applicable.
        
        Parameters
        ----------
        force : `bool`
            Whether the transport should be shut down immediately.
        """
        transport = self.get_transport()
        if (transport is not None):
            transport.close()
            
            if force:
                transport.abort()
    
    
    def get_transport(self):
        """
        Returns the transport layer of the protocol.
        
        Returns
        -------
        transport : `None`, ``AbstractTransportLayerBase``
        """
        pass
    
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any` = `None`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return default
    
    
    def set_exception(self, exception):
        """
        Called by ``.connection_lost`` if the connection is closed by an exception, or can be called if any method
        of the protocol raises an unexpected exception.
        
        Parameters
        ----------
        exception : `BaseException`
        """
        pass
    
    
    def eof_received(self):
        """
        Calling ``.connection_lost`` without exception causes eof.
        
        Marks the protocol as it is at eof and stops payload processing if applicable.
        
        Returns
        -------
        transport_closes : `bool`
            Returns `False` if the transport will close itself. If it returns `True`, then closing the transport is up
            to the protocol.
            
            Always returns `False`.
        """
        pass
    
    
    def data_received(self, data):
        """
        Called when some data is received.
        
        Parameters
        ----------
        data : `bytes`
            The received data.
        """
        pass
    
    
    def pause_writing(self):
        """
        Called when the transport's buffer goes over the high-water mark.
        
        ``.pause_writing`` is called when the buffer goes over the high-water mark, and eventually
        ``.resume_writing`` is called when the buffer size reaches the low-water mark.
        """
        pass
    
    
    def resume_writing(self):
        """
        Called when the transport's buffer drains below the low-water mark.
        
        See ``.pause_writing`` for details.
        """
        pass
    
    
    def write(self, data):
        """
        Writes the given data to the protocol's transport.
        
        Parameters
        ----------
        data : `bytes-like`
            The data to write.
        
        Raises
        ------
        RuntimeError
            Protocol has no attached transport.
        """
        pass
    
    
    def writelines(self, lines):
        """
        Writes the given lines to the protocol's transport.
        
        Parameters
        ----------
        lines : `iterable` of `bytes-like`
            The lines to write.
        
        Raises
        ------
        RuntimeError
            Protocol has no attached transport.
        """
        pass
    
    
    def write_eof(self):
        """
        Writes eof to the transport's protocol if applicable.
        """
        pass
    
    
    def can_write_eof(self):
        """
        Returns whether ``.write_eof`` will write eof successfully.
        
        Returns
        -------
        can_write_eof : `bool`
        """
        pass
    
    
    async def drain(self):
        """
        Blocks till the write buffer is drained.
        
        This method is a coroutine.
        
        Raises
        ------
        BaseException
            Connection lost exception if applicable.
        """
        pass
    
    
    def datagram_received(self, data, address):
        """
        Called when some datagram is received.
        
        Parameters
        ----------
        data : `bytes`
            The received data.
        address : `tuple` (`str`, `int`)
            The address from where the data was received.
        """
        pass
    
    
    def error_received(self, exception):
        """
        Called when a send or receive operation raises an `OSError`, but not `BlockingIOError`, nor `InterruptedError`.
        
        Parameters
        ----------
        exception : `OSError`
            The catched exception.
        """
        pass


class AbstractBidirectionalTransportLayerBase(AbstractTransportLayerBase, AbstractProtocolBase):
    """
    Defines abstract bidirectional transport layer functionality.
    """
    __slots__ = ()



__all__ = ('IgnoreCaseString', )

import sys

from .docs import has_docs


@has_docs
class IgnoreCaseString(str):
    """
    Strings, which have their casing ignored.
    
    Attributes
    ----------
    _case_fold : `str`
        Case folded version of the string.
    """
    __slots__ = ('_case_fold', )
    
    @has_docs
    def __new__(cls, value='', encoding=sys.getdefaultencoding(), errors='strict'):
        """
        Return an string which ignores casing. If object is not provided, returns the empty string. Otherwise, the
        behavior of ``IgnoreCaseString`` depends on whether encoding or errors is given, as follows.
        
        If neither encoding nor errors is given, `IgnoreCaseString(object)` returns `object.__str__()`, which is the
        "informal" or nicely printable string representation of object. For string objects, this is `string`. If
        object does not have a `__str__()` method, then `IgnoreCaseString()` falls back to returning `repr(object)`.
        
        If at least one of encoding or errors is given, object should be a `bytes-like` object. In this case, if object
        is a `byte-like` object, then `IgnoreCaseString(bytes, encoding, errors)` is equivalent to
        `bytes.decode(encoding, errors)`. Otherwise, the bytes object underlying the buffer object is obtained before
        calling `bytes.decode()`.
        
        Passing a `bytes-like` object to ``IgnoreCaseString`` without the encoding or errors parameters falls under the
        first case of returning the informal string representation.
        
        Parameters
        ----------
        value : `Any` = `''`, Optional
            The value, what is representation or encoded version is returned.
        encoding : `str` = `sys.getdefaultencoding()`, Optional
            Encoding to use when decoding a `bytes-like`.
        errors : `str` = `'strict'`, Optional
            May be given to set a different error handling scheme when decoding from `bytes-like`. The default `errors`
            value is `'strict'`, meaning that encoding errors raise a `UnicodeError`. Other possible values are
            `'ignore'`, `'replace'`, `'xmlcharrefreplace'`, `'backslashreplace'` and any other name registered via
            `codecs.register_error()`.
        
        Returns
        -------
        self : ``IgnoreCaseString``
        """
        if type(value) is cls:
            return value
        
        if isinstance(value,(bytes, bytearray, memoryview)):
            value = str(value, encoding, errors)
        elif isinstance(value, str):
            pass
        else:
            value = str(value)
        
        self = str.__new__(cls, value)
        self._case_fold = str.casefold(value)
        return self
    
    @has_docs
    def __hash__(self):
        """Returns the string's hash value."""
        return hash(self._case_fold)
    
    @has_docs
    def __eq__(self, other):
        """Returns whether the two strings are equal."""
        other_type = other.__class__
        if other_type is type(self):
            other_value = other._case_fold
        elif other_type is str:
            other_value = other.casefold()
        elif issubclass(other, str):
            other_value = str.casefold(other)
        else:
            return NotImplemented
        
        return (self._case_fold == other_value)



__all__ = ()

class ConnectionKey:
    """
    Contains information about a host, like proxy, TLS to prevent reusing wrong connections from the pool.
    
    Attributes
    ----------
    host : `str`
        The host's ip address.
    is_ssl : `bool`
        Whether the connection is secure.
    port : `int`
        The host's port.
    proxy_auth : `None`, ``BasicAuth``
        Proxy authorization.
    proxy_url : `None`, ``URL``
        Proxy's url.
    ssl : `None`, ``SSLContext``, `bool`, ``Fingerprint``
        The connection's ssl type.
    """
    __slots__ = ('host', 'is_ssl', 'port', 'proxy_auth',  'proxy_url', 'ssl',) # + 'proxy_header_hash',
    
    def __init__(self, request):
        # proxy_headers = request.proxy_headers
        # if request.proxy_headers is not None:
        #     proxy_header_hash = hash(tuple(proxy_headers.items()))
        # else:
        #     proxy_header_hash = None
        
        self.host = request.host
        self.port = request.port
        self.is_ssl = request.is_ssl()
        self.ssl = request.ssl
        self.proxy_auth = request.proxy_auth
        self.proxy_url = request.proxy_url
        # self.proxy_header_hash = proxy_header_hash
    
    def __repr__(self):
        """Returns the connection key's representation."""
        return f'<{self.__class__.__name__} host={self.host!r}, port={self.port!r}>'
    
    def __eq__(self, other):
        """Returns whether the two connection keys are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.host != other.host:
            return False
        
        if self.port != other.port:
            return False
        
        if self.is_ssl != other.is_ssl:
            return False
        
        if self.ssl is None:
            if other.ssl is not None:
                return False
        else:
            if other.ssl is None:
                return False
            
            if self.ssl != other.ssl:
                return False
            
        if self.proxy_auth is None:
            if other.proxy_auth is not None:
                return False
        else:
            if other.proxy_auth is None:
                return False
            
            if self.proxy_auth != other.proxy_auth:
                return False

        if self.proxy_url is None:
            if other.proxy_url is not None:
                return False
        else:
            if other.proxy_url is None:
                return False
            
            if self.proxy_url != other.proxy_url:
                return False
        
        return True
    
    def __hash__(self):
        """Returns the connection key's hash value."""
        return hash(self.host) ^ (self.port << 17) ^ hash(self.is_ssl) ^ hash(self.ssl) ^ hash(self.proxy_auth) ^ \
               hash(self.proxy_url)



__all__ = ()


class _HandleCancellerBase:
    """
    ``Future`` callback-base to cancel a respective ``Handle`` when the future is marked as done before the
    handle, or if the handle runs first, then sets the future result or exception.
    
    This class do not implements ``.__call__`. Subclasses should implement that.
    
    Attributes
    ----------
    _handle : `None`, ``Handle``
        The handle to cancel, when the future is marked as done before the respective handle ran.
    """
    __slots__ = ('_handle',)
    
    def __init__(self):
        """
        Creates a new handle canceller.
        
        Note, that a handle canceller is created with empty ``.handle``, because it is a callback and a ``Handle``'s
        function at the same time, so ``.handle`` is needed to be set from outside.
        """
        self._handle = None
    
    def __call__(self, future):
        """
        Called by the respective ``Future`` as callback, or by the respective handle.
        
        Subclasses should overwrite this method.
        
        Parameters
        ----------
        future : ``Future``
            The respective future to what the handle canceller was added as callback.
        """
        pass
    
    def cancel(self):
        """
        Cancels the handle.
        
        This method is usually called when the respective future's handles are cleared.
        """
        handle = self._handle
        if handle is None:
            return
         
        self._handle = None
        handle.cancel()
    
    
    def __repr__(self):
        """Returns the handler canceller's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        handle = self._handle
        if handle is None:
            state = 'cancelled'
        else:
            state = 'alive'
        
        repr_parts.append(' ')
        repr_parts.append(state)
        
        if (handle is not None):
            repr_parts.append(' handle=')
            repr_parts.append(handle._get_repr_without(self))
        
        repr_parts.append('>')
        return ''.join(repr_parts)


class _SleepHandleCanceller(_HandleCancellerBase):
    """
    ``Future`` callback to cancel a respective ``Handle`` when the future is marked as done before the
    handle, or if the handle runs first, then sets the future result as `None˛.
    
    Attributes
    ----------
    _handle : `None`, ``Handle``
        The handle to cancel, when the future is marked as done before the respective handle ran.
    """
    __slots__ = ()
    
    def __call__(self, future):
        """
        Called by the respective ``Future`` as callback, or by the respective handle. Sets the given
        `future`'s result as `None` if applicable.
        
        Sets ``._handle`` as `None`, marking the canceller as it ran already.
        
        Parameters
        ----------
        future : ``Future``
            The respective future to what the handle canceller was added as callback.
        """
        handle = self._handle
        if handle is None:
            return
        
        self._handle = None
        handle.cancel()
        future.set_result_if_pending(None)


class _TimeoutHandleCanceller(_HandleCancellerBase):
    """
    ``Future`` callback to cancel a respective ``Handle`` when the future is marked as done before the
    handle, or if the handle runs first, then sets the future's exception to `TimeoutError`.
    
    Attributes
    ----------
    _handle : `None`, ``Handle``
        The handle to cancel, when the future is marked as done before the respective handle ran.
    """
    __slots__ = ()
    
    def __call__(self, future):
        """
        Called by the respective ``Future`` as callback, or by the respective handle. Sets the given
        `future`'s exception to `TimeoutError` if applicable.
        
        Sets ``._handle`` as `None`, marking the canceller as it ran already.
        
        Parameters
        ----------
        future : ``Future``
            The respective future to what the handle canceller was added as callback.
        """
        handle = self._handle
        if handle is None:
            return
        
        handle.cancel()
        self._handle = None
        future.set_exception_if_pending(TimeoutError())



__all__ = ('Lock', 'ScarletLock')

import warnings
from collections import deque

from .future import Future


class Lock:
    """
    Implements a mutex lock for hata tasks.

    A hata lock can be used to guarantee exclusive access to a shared resource.
    
    The preferred way of using a hata lock is with `async with` statement:
    
    ```py
    lock = Lock(loop)
    
    async with lock:
        # access resources
    ```
    
    You can also manually acquire and release locks, like:
    
    ```py
    lock = Lock(loop)
    
    await lock.acquire()
    try:
        # access resources
    finally:
        lock.release()
    ```
    
    Attributes
    ----------
    _loop : ``EventThread``
        The event loop to what the lock is bound to.
    _waiters : `deque` of ``Future``
        Futures on which the suspended tasks wait.
    """
    __slots__ = ('_loop', '_waiters', )
    
    def __new__(cls, loop):
        """
        Creates a new lock instance.
        
        loop : ``EventThread``
            The event loop to what the lock will be bound to.
        """
        self = object.__new__(cls)
        self._loop = loop
        self._waiters = deque()
        return self
    
    async def __aenter__(self):
        """
        Acquires the lock.
        
        This method is a coroutine.
        """
        future = Future(self._loop)
        waiters = self._waiters
        waiters.appendleft(future)
        
        if len(waiters) > 1:
            waiter = waiters[1]
            try:
                await waiter
            except:
                try:
                    waiters.remove(waiter)
                except ValueError:
                    pass
                
                raise
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Releases the lock.
        
        This method is a coroutine.
        """
        future = self._waiters.pop()
        future.set_result_if_pending(None)
        return False
    
    
    def locked(self):
        warnings.warn(
            f'{self.__class__.__name__}.locked` is deprecated.',
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_locked()
    
    
    def is_locked(self):
        """
        Returns whether the lock is entered anywhere.
        
        Returns
        -------
        is_locked: `bool`
        """
        if self._waiters:
            return True
        
        return False
    

    def __iter__(self):
        """
        Blocks until all the lock is unlocked everywhere. If lock is used meanwhile anywhere meanwhile the we acquire
        it, will await that as well.
        
        This method is a generator. Should be used with `await` expression.
        """
        waiters = self._waiters
        while waiters:
            yield from waiters[0]
    
    __await__ = __iter__
    
    acquire = __aenter__
    
    def release(self):
        """Releases the lock."""
        future = self._waiters.pop()
        future.set_result_if_pending(None)
    
    def __repr__(self):
        """Returns the lock's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' locked=',
        ]
        
        count = len(self._waiters)
        if count:
            repr_parts.append('True, waiting=')
            repr_parts.append(repr(count))
        else:
            repr_parts.append('False')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)


class ScarletLock(Lock):
    """
    A hata scarlet lock can be used to guarantee access to a shared resource `n` amount of times.
    
    Should be used with `async with` statement
    
    Attributes
    ----------
    _loop : ``EventThread``
        The event loop to what the lock is bound to.
    _waiters : `deque` of ``Future``
        Futures on which the suspended tasks wait.
    _size : `int`
        The maximal amount of parallel entries to this lock.
    """
    __slots__ = ('_size',)
    
    def __new__(cls, loop, size=1):
        """
        Creates a new lock instance.
        
        loop : ``EventThread``
            The event loop to what the lock will be bound to.
        size : `int`
            The maximal amount of parallel entries to this lock.
        
        Raises
        ------
        TypeError
            `size` is not given as `int`.
        ValueError
            `size` is given as non negative `int`.
        """
        size_type = size.__class__
        if size_type is int:
            pass
        elif issubclass(size_type, int):
            size = int(size)
        else:
            raise TypeError(
                f'`size` can be `int`, got {size_type.__name__}; {size!r}.'
            )
        
        if size < 1:
            raise ValueError(
                f'`size` can be only positive, got {size!r}.'
            )
        
        self = object.__new__(cls)
        self._loop = loop
        self._waiters = deque()
        self._size = size
        
        return self
    
    async def __aenter__(self):
        """
        Acquires the lock.
        
        This method is a coroutine.
        """
        future = Future(self._loop)
        waiters = self._waiters
        waiters.appendleft(future)
        
        size = self._size
        if len(waiters) > size:
            waiter = waiters[size]
            try:
                await waiter
            except:
                try:
                    waiters.remove(waiter)
                except ValueError:
                    pass
                
                raise
    
    acquire = __aenter__
    
    def get_size(self):
        """
        Returns the size of the ``ScarletLock``
        
        Returns
        -------
        size : `int`
        """
        return self._size
    
    
    def get_acquired(self):
        """
        Returns how much times the lock is acquired currently. Caps at the size of the lock.
        
        Returns
        --------
        acquired : `int`
        """
        waiter_count = len(self._waiters)
        size = self._size
        if waiter_count > size:
            waiter_count = size
        
        return waiter_count
    
    
    def get_waiting(self):
        """
        Returns how tasks are waiting to acquire the lock.
        
        Returns
        --------
        waiting : `int`
        """
        waiter_count = len(self._waiters)
        size = self._size
        if waiter_count > size:
            waiting = waiter_count - size
        else:
            waiting = 0
        
        return waiting
    
    # returns True if the Lock is entered anywhere
    def is_locked(self):
        """
        Returns whether the lock is entered same or more times than ``.size`` is set to.
        
        Returns
        -------
        is_locked: `bool`
        """
        if len(self._waiters) >= self._size:
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the scarlet lock's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' size=',
        ]
        
        size = self._size
        repr_parts.append(repr(size))
        
        repr_parts.append(', locked=')
        count = len(self._waiters)
        if count >= size:
            repr_parts.append('True')
        else:
            repr_parts.append('False')
        
        if count:
            repr_parts.append(', waiting=')
            repr_parts.append(repr(count))
        
        repr_parts.append('>')
        return ''.join(repr_parts)



__all__ = ()

import sys
from selectors import DefaultSelector
from threading import current_thread


if sys.platform == 'win32':
    # If windows select raises OSError, we cannot do anything, but if it it raises ValueError, we can increases
    # windows select() from 500~ till the hard limit, with sharding up it's polls
    
    from select import select
    MAX_FD_S = 500 #512 is the actual amount?
    MAX_SLEEP = 0.001
    EMPTY = []
    
    class DefaultSelector(DefaultSelector):
        """
        Selector subclass for windows to bypass default limit.
        
        Note, that this selector might become CPU heavy if the limit is passed and the sockets might become closed
        if too much is open.
        
        I do not take credit for any misbehaviour.
        """
        def _select(self, r, w, _, timeout=None):
            try:
                result_r, result_w, result_x = select(r, w, w, timeout)
            except ValueError:
                default_reader = current_thread()._self_read_socket
                r.remove(default_reader.fileno())
                
                sharded_r = []
                sharded_w = []
                
                sharded = [(sharded_r, sharded_w,),]
                
                count = 0
                for reader in r:
                    if count == MAX_FD_S:
                        sharded_r = [reader]
                        sharded_w = []
                        sharded.append((sharded_r, sharded_w),)
                        count = 1
                    else:
                        sharded_r.append(reader)
                        count = count + 1
                
                for writer in w:
                    if count == MAX_FD_S:
                        sharded_r = []
                        sharded_w = [writer]
                        sharded.append((sharded_r, sharded_w),)
                        count = 1
                    else:
                        sharded_w.append(writer)
                        count += 1
                
                collected_r = []
                collected_w = []
                
                r.add(default_reader.fileno())
                
                for iter_r, iter_w in sharded:
                    try:
                        result_r, result_w, result_x = select(iter_r, iter_w, iter_w, 0.0)
                    except OSError:
                        remove = []
                        for reader in iter_r:
                            try:
                                l = [reader]
                                result_r, result_w, result_x = select(l, EMPTY, EMPTY, 0.0)
                            except OSError:
                                remove.append(reader)
                            else:
                                if result_r:
                                    collected_r.append(result_r[0])
                        
                        if remove:
                            for reader in remove:
                                r.discard(reader)
                                
                                try:
                                    self.unregister(reader)
                                except KeyError:
                                    pass
                            remove.clear()
                        
                        for writer in iter_w:
                            try:
                                l = [writer]
                                result_r, result_w, result_x=select(EMPTY, l, l, 0.0)
                            except OSError:
                                remove.append(writer)
                            else:
                                if result_w:
                                    collected_w.append(result_w[0])
                                elif result_x:
                                    collected_w.append(result_x[0])
                        
                        if remove:
                            for writer in remove:
                                w.discard(writer)
                                
                                try:
                                    self.unregister(writer)
                                except KeyError:
                                    pass
                            remove.clear()
                    else:
                        collected_r.extend(result_r)
                        collected_w.extend(result_w)
                        collected_w.extend(result_x)

                if (not collected_r) and (not collected_w):
                    if timeout is None:
                        timeout = MAX_SLEEP
                    elif timeout < 0.0:
                        timeout = 0.0
                    elif timeout > MAX_SLEEP:
                        timeout = MAX_SLEEP
                    
                    result_r, result_w, result_x = select([default_reader], EMPTY, EMPTY, timeout)
                    collected_r.extend(result_r)
                
                return collected_r, collected_w, EMPTY
            
            except OSError:
                collected_r = []
                collected_w = []
                do_later_r = []
                do_later_w = []
                remove = []
                for reader in r:
                    try:
                        l = [reader]
                        result_r, result_w, result_x = select(l, EMPTY, EMPTY, 0.0)
                    except OSError:
                        remove.append(reader)
                    else:
                        if result_r:
                            collected_r.append(result_r[0])
                        else:
                            do_later_r.append(reader)
                
                if remove:
                    for reader in remove:
                        r.discard(reader)
                        
                        try:
                            self.unregister(reader)
                        except KeyError:
                            pass
                    remove.clear()
                    
                for writer in w:
                    try:
                        l = [writer]
                        result_r, result_w, result_x = select(EMPTY, l, l, 0.0)
                    except OSError:
                        remove.append(writer)
                    else:
                        if result_w:
                            collected_w.append(result_w[0])
                        elif result_x:
                            collected_w.append(result_x[0])
                        else:
                            do_later_w.append(writer)
                    
                if remove:
                    for writer in remove:
                        w.discard(writer)
                        
                        try:
                            self.unregister(writer)
                        except KeyError:
                            pass
                    remove.clear()
                
                if collected_r or collected_w:
                    return collected_r, collected_w, EMPTY
                
                result_r, result_w, result_x = select(r, w, w, timeout)
                result_w.extend(result_x)
                return result_r, result_w, EMPTY
            else:
                result_w.extend(result_x)
                return result_r, result_w, EMPTY



__all__ = ('JSONDecodeError', 'from_json', 'to_json')

from enum import Enum
from json import JSONDecodeError, dumps as dump_to_json, loads as from_json

from .docs import has_docs


@has_docs
def added_json_serializer(obj):
    """
    Default json encoder function for supporting additional object types.
    
    Parameters
    ----------
    obj : `iterable`, `Enum`
        The value to json serialise.
    
    Returns
    -------
    result : `Any`
    
    Raises
    ------
    TypeError
        If the given object is not json serializable.
    """
    if hasattr(obj, '__iter__'):
        return list(obj)
    
    if isinstance(obj, Enum):
        return obj.value
    
    raise TypeError(f'Object of type {obj.__class__.__name__!r} is not JSON serializable, got {obj!r}')


@has_docs
def to_json(data):
    """
    Converts the given object to json.
    
    Parameters
    ----------
    data : `Any`
    
    Returns
    -------
    json : `str`
    
    Raises
    ------
    TypeError
        If the given object is /or contains an object with a non convertible type.
    """
    return dump_to_json(data, separators=(',', ':'), ensure_ascii=True, default=added_json_serializer)



__all__ = ('cached_property', 'class_property', 'docs_property', 'module_property', 'name_property',)

from .docs import has_docs


NoneType = type(None)


@has_docs
class docs_property:
    """
    Property to return the class's docs if called from class, else the given object.
    """
    __slots__ = ()
    
    @has_docs
    def __init__(self):
        """
        Creates a new docs property.
        """
    
    def __get__(self, obj, type_):
        if obj is None:
            return type_.__class_doc__
        else:
            return obj.__instance_doc__
    
    def __set__(self, obj, value):
        raise AttributeError('can\'t set attribute')
    
    def __delete__(self, obj):
        raise AttributeError('can\'t delete attribute')


@has_docs
class name_property:
    """
    Property to return the class's name if called from the respective class.
    
    Attributes
    ----------
    class_name : `str`
        The class's name.
    fget : `callable`
        Callable what's return will be returned, when called from an instance.
    """
    __slots__ = ('class_name', 'fget')
    
    @has_docs
    def __init__(self, name, fget):
        """
        Creates a new docs property.
        """
        self.class_name = name
        self.fget = fget
    
    def __get__(self, obj, type_):
        if obj is None:
            return type_.class_name
        else:
            return self.fget(obj)
    
    def __set__(self, obj, value):
        raise AttributeError('can\'t set attribute')
    
    def __delete__(self, obj):
        raise AttributeError('can\'t delete attribute')


@has_docs
class module_property:
    """
    Instead of defining a  `.__module__` attribute as `property`, define it as `module_property` to avoid getter issues,
    when calling from class.
    
    Attributes
    ----------
    fget : `func`
        Getter used when calling the module from instance.
    module : `str`
        The module where the `module_property` was created.
    """
    __slots__ = ('fget', 'module', )
    
    def __new__(cls, fget):
        module = getattr(fget, '__module__', None)
        if module is None:
            module = cls.__module__
        elif type(module) is str:
            module = module
        elif isinstance(module, str):
            module = str(module)
        else:
            module = cls.__module__
        
        self = object.__new__(cls)
        self.fget = fget
        self.module = module
        return self
    
    def __get__(self, obj, type_):
        if obj is None:
            return self.module
        
        return self.fget()
    
    def __set__(self, obj, module):
        if module is None:
            module = self.__module__
        elif type(module) is str:
            module = module
        elif isinstance(module, str):
            module = str(module)
        else:
            module = self.__module__
        
        self.module = module
    
    def __delete__(self, obj):
        raise AttributeError('can\'t delete attribute')


class class_property:
    __doc__ = docs_property()
    
    __class_doc__ = ("""
    Class level property.
    
    Attributes
    ----------
    fget : `callable`
        getter method.
    fset : `callable`
        Setter method.
    fdel : `callable`
        Deleter method.
    __instance_doc__ : `Any`
        Documentation for the property.
    """)
    
    def __new__(cls, fget=None, fset=None, fdel=None, doc=None):
        """
        Creates a new ``class_property`` from the given parameters.
        
        If `doc` is not given or given as `None`, it will default to `fget`'s if applicable.
        
        Parameters
        ----------
        fget : `None`, `callable` = `None`, Optional
            getter method.
        fset : `None`, `callable` = `None`, Optional
            Setter method.
        fdel : `None`, `callable` = `None`, Optional
            Deleter method.
        doc : `None`, `Any` = `None`, Optional
            Documentation for the property.
        """
        if (doc is None) and (fget is not None):
            doc = fget.__doc__
        
        self = object.__new__(cls)
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        self.__instance_doc__ = doc
        return self
    
    def __get__(self, obj, type_):
        getter = self.fget
        
        if getter is None:
            raise AttributeError('unreadable attribute')
        
        return getter(type_)
    
    def __set__(self, obj, value):
        setter = self.fset
        if setter is None:
            raise AttributeError('can\'t set attribute')
        
        return setter(type(obj), value)
    
    def __delete__(self, obj):
        deleter = self.fdel
        if deleter is None:
            raise AttributeError('can\'t delete attribute')
        
        return deleter(type(obj))
    
    def getter(self, fget):
        """
        Returns a new property with getter set.
        
        Parameters
        ----------
        fget : `callable`
            Getter method.
        
        Returns
        -------
        new : ``class_property``
        """
        return type(self)(fget, self.fset, self.fdel, self.__instance_doc__)
    
    def setter(self, fset):
        """
        Returns a new property with setter set.
        
        Parameters
        ----------
        fset : `callable`
            Setter method.
        
        Returns
        -------
        new : ``class_property``
        """
        return type(self)(self.fget, fset, self.fdel, self.__instance_doc__)
    
    def deleter(self, fdel):
        """
        Returns a new property with deleter set.
        
        Parameters
        ----------
        fdel : `callable`
            Deleter method.
        
        Returns
        -------
        new : ``class_property``
        """
        return type(self)(self.fget, self.fset, fdel, self.__instance_doc__)


@has_docs
class cached_property:
    """
    Cached property, what can be used as a method decorator. It operates almost like python's `@property`, but it puts
    the result of the method to the respective instance's `_cache`, preferably type `dict` attribute.
    
    Attributes
    ----------
    fget : `function`
        Getter method of the cached property.
    name : `str`
        The name of the cached property.
    """
    __slots__ = ('fget', 'name',)
    
    @has_docs
    def __new__(cls, fget):
        """
        Creates a new cached property instance with the given getter.
        
        Parameters
        ----------
        fget : `function`
            Getter method of the cached property.

        Raises
        ------
        TypeError
            If `fget` has no name or it's name is not `str`.
        """
        name = getattr(fget, '__name__', None)
        
        name_type = name.__class__
        if name_type is NoneType:
            name_type_correct = False
        elif name_type is str:
            name_type_correct = True
        elif issubclass(name_type, str):
            name = str(name)
            name_type_correct = True
        else:
            name_type_correct = False
        
        if not name_type_correct:
            raise TypeError(
                f'`fget.__name__` can be `str`, got {name_type.__name__}; {name!r}.'
            )
        
        self = object.__new__(cls)
        self.fget = fget
        self.name = name
        return self
    
    def __get__(self, obj, type_):
        if obj is None:
            return self
        
        value = obj._cache.get(self.name, ...)
        if value is ...:
            value = self.fget(obj)
            obj._cache[self.name] = value
        
        return value
    
    def __set__(self, obj, value):
        raise AttributeError('can\'t set attribute')
    
    def __delete__(self, obj):
        raise AttributeError('can\'t delete attribute')



__all__ = ('WeakItemDictionary',)

from .dict_iterator_bases import DictionaryItemIteratorBase, DictionaryValueIteratorBase
from .dict_update_iterable_iterator import _dict_update_iterable_iterator
from .docs import copy_docs, has_docs
from .weak_core import KeyedReferer, WeakReferer, add_to_pending_removals
from .weak_key_dictionary import _WeakKeyDictionaryCallback, _WeakKeyDictionaryKeyIterator


@has_docs
class _WeakItemDictionaryValueCallback:
    """
    Callback used by ``WeakItemDictionary``-s.
    
    Attributes
    ----------
    _parent : ``WeakReferer`` to ``WeakItemDictionary``
        The parent weak or hybrid value dictionary.
    """
    __slots__ = ('_parent', )
    
    @has_docs
    def __new__(cls, parent):
        """
        Creates a new ``_WeakItemDictionaryValueCallback``.
        
        Parameters
        ----------
        parent : ``WeakItemDictionary``
            The parent weak or hybrid value dictionary.
        """
        parent = WeakReferer(parent)
        self = object.__new__(cls)
        self._parent = parent
        return self
    
    
    @has_docs
    def __call__(self, reference):
        """
        Called when a value of the respective weak value is garbage collected.
        
        Parameters
        ----------
        reference : ``KeyedReferer``
            Weakreference to the respective object.
        """
        parent = self._parent()
        if parent is None:
            return
        
        key_reference = reference.key
        
        if parent._iterating:
            add_to_pending_removals(parent, key_reference)
        else:
            try:
                dict.__delitem__(parent, key_reference)
            except KeyError:
                pass


class _WeakItemDictionaryValueIterator(DictionaryValueIteratorBase):
    __slots__ = ()
    
    @copy_docs(DictionaryValueIteratorBase.__iter__)
    def __iter__(self):
        parent = self._parent
        parent._iterating += 1
        
        try:
            for value_reference in dict.values(parent):
                value = value_reference()
                if (value is None):
                    add_to_pending_removals(parent, value_reference.key)
                    continue
                
                yield value
                continue
        
        finally:
            parent._iterating -= 1
            parent._commit_removals()
    
    
    @copy_docs(DictionaryValueIteratorBase.__contains__)
    def __contains__(self, contains_value):
        parent = self._parent
        for value_reference in dict.values(parent):
            value = value_reference()
            if (value is None):
                add_to_pending_removals(parent, value_reference.key)
                continue
            
            if value == contains_value:
                result = True
                break
        
        else:
            result = False
        
        parent._commit_removals()
        
        return result
    
    

class _WeakItemDictionaryItemIterator(DictionaryItemIteratorBase):
    __slots__ = ()
    
    @copy_docs(DictionaryItemIteratorBase.__iter__)
    def __iter__(self):
        parent = self._parent
        iterating = parent._iterating
        parent._iterating = iterating + 1
        
        try:
            for key_reference, value_reference in dict.items(parent):
                key = key_reference()
                if (key is None):
                    add_to_pending_removals(parent, key_reference)
                    continue
                
                value = value_reference()
                if (value is None):
                    add_to_pending_removals(parent, key_reference)
                    continue
                
                yield key, value
                continue
        
        finally:
            parent._iterating -= 1
            parent._commit_removals()
    
    
    @copy_docs(DictionaryItemIteratorBase.__contains__)
    def __contains__(self, contains_item):
        if not isinstance(contains_item, tuple):
            return False
        
        if len(contains_item) != 2:
            return False
        
        parent = self._parent
        
        contains_key, contains_value = contains_item
        
        try:
            contains_key_reference = WeakReferer(contains_key)
        except TypeError:
            return False
        
        value_reference = dict.get(self._parent, contains_key_reference)
        if value_reference is None:
            return False
        
        value = value_reference()
        if value is None:
            if parent._iterating:
                add_to_pending_removals(parent, value_reference.key)
            else:
                dict.__delitem__(parent, value_reference.key)
            
            return False
        
        return (value == contains_value)


@has_docs
class WeakItemDictionary(dict):
    """
    weak item dictionary, which stores it's values weakly referenced.
    
    Attributes
    ----------
    _pending_removals : `None`, `set` of (``KeyedReferer``, ``WeakHasher``)
        Pending removals of the weak item dictionary if applicable.
    _iterating : `int`
        Whether the weak item dictionary is iterating and how much times.
    _key_callback : ``_WeakKeyDictionaryCallback``
        Callback added to the ``WeakItemDictionary``'s weak keys.
    _value_callback : ``_WeakItemDictionaryValueCallback``
        Callback added to the ``WeakItemDictionary``'s weak values.
    
    Class Attributes
    ----------------
    MAX_REPR_ELEMENT_LIMIT : `int` = `50`
        The maximal amount of items to render by ``.__repr__``.
    
    Notes
    -----
    ``WeakItemDictionary``-s are weakreferable.
    """
    __slots__ = ('__weakref__', '_pending_removals', '_iterating', '_key_callback', '_value_callback')
    
    MAX_REPR_ELEMENT_LIMIT = 50
    
    @has_docs
    def _commit_removals(self):
        """
        Commits the pending removals of the weak item dictionary if applicable.
        """
        if self._iterating:
            return
        
        pending_removals = self._pending_removals
        if pending_removals is None:
            return
        
        self._pending_removals = None
        
        for key_reference in pending_removals:
            try:
                dict.__delitem__(self, key_reference)
            except KeyError:
                pass
    
    # __class__ -> same
    
    @has_docs
    def __contains__(self, key):
        """Returns whether the weak item dictionary contains the given key."""
        try:
            key = WeakReferer(key)
        except TypeError:
            return False
        
        value_reference = dict.get(self, key, None)
        if value_reference is None:
            return False
        
        value = value_reference()
        if (value is not None):
            return True
        
        if self._iterating:
            add_to_pending_removals(self, value_reference.key)
        else:
            dict.__delitem__(self, key)
        
        return False
    
    # __delattr__ -> same
    
    @has_docs
    def __delitem__(self, key):
        """Deletes the value of the weak item dictionary which matches the given key."""
        try:
            key = WeakReferer(key)
        except TypeError:
            raise KeyError(key) from None
        
        dict.__delitem__(self, key)
    
    # __dir__ -> same
    # __doc__ -> same
    
    def __eq__(self, other):
        """Returns whether the two dictionaries are the same."""
        if isinstance(other, type(self)):
            return dict.__eq__(self, other)
        
        if isinstance(other, dict):
            pass
        
        elif hasattr(type(other), '__iter__'):
            try:
                other = dict(other)
            except (ValueError, TypeError):
                return NotImplemented
        
        else:
            return NotImplemented
        
        self_dict = dict(self.items())
        
        return self_dict == other
    
    # __format__ -> same
    # __ge__ -> same
    # __getattribute__ -> same
    
    @has_docs
    def __getitem__(self, key):
        """Gets the value of the weak item dictionary which matches the given key."""
        try:
            key_reference = WeakReferer(key)
        except TypeError:
            raise KeyError(key) from None
        
        value_reference = dict.__getitem__(self, key_reference)
        value = value_reference()
        if (value is not None):
            return value
        
        if self._iterating:
            add_to_pending_removals(self, value_reference.key)
        else:
            dict.__delitem__(self, value_reference.key)
        
        raise KeyError(key)
    
    # __gt__ -> same
    # __hash__ -> same
    
    @has_docs
    def __init__(self, iterable=None):
        """
        Creates a new ``WeakItemDictionary`` from the given iterable.
        
        Parameters
        ----------
        iterable : `None`, `iterable` = `None`, Optional
            Iterable to update the created dictionary with.
        """
        self._pending_removals = None
        self._iterating = 0
        self._key_callback = _WeakKeyDictionaryCallback(self)
        self._value_callback = _WeakItemDictionaryValueCallback(self)
        if (iterable is not None):
            self.update(iterable)
    
    # __init_subclass__ -> same
    
    @has_docs
    def __iter__(self):
        """Returns a ``_WeakItemDictionaryKeyIterator`` iterating over the weak item dictionary's keys."""
        return iter(_WeakKeyDictionaryKeyIterator(self))
    
    # __le__ -> same
    
    @has_docs
    def __len__(self):
        """Returns the length of the weak item dictionary."""
        length = dict.__len__(self)
        pending_removals = self._pending_removals
        if (pending_removals is not None):
            length -= len(pending_removals)
        
        return length
    
    # __lt__ -> same
    
    def __ne__(self, other):
        """Returns whether the two dictionaries are different."""
        if isinstance(other, type(self)):
            return dict.__ne__(self, other)
        
        if isinstance(other, dict):
            pass
        
        elif hasattr(type(other), '__iter__'):
            try:
                other = dict(other)
            except (ValueError, TypeError):
                return NotImplemented
        
        else:
            return NotImplemented
        
        self_dict = dict(self.items())
        
        return self_dict != other
    
    # __new__ -> same
    
    @has_docs
    def __reduce__(self):
        """Reduces the dictionary to a picklable object."""
        return (type(self), list(self.items()))
    
    
    @has_docs
    def __reduce_ex__(self, version):
        """Reduces the dictionary to a picklable object."""
        return type(self).__reduce__(self)
    
    
    @has_docs
    def __repr__(self):
        """Returns the representation of the weak item dictionary."""
        result = [self.__class__.__name__, '({']
        
        if len(self):
            limit = self.MAX_REPR_ELEMENT_LIMIT
            collected = 0
            for key_reference, value_reference in dict.items(self):
                key = key_reference()
                if (key is None):
                    add_to_pending_removals(self, key_reference)
                    continue
                
                value = value_reference()
                if (value is None):
                    add_to_pending_removals(self, key_reference)
                    continue
                
                result.append(repr(key))
                result.append(': ')
                result.append(repr(value))
                result.append(', ')
                
                collected +=1
                if collected != limit:
                    continue
                
                leftover = len(self) - collected
                if leftover:
                    result.append('...}, ')
                    result.append(str(leftover))
                    result.append(' truncated)')
                else:
                    result[-1] = '})'
                break
            else:
                result[-1] = '})'
            
            self._commit_removals()
        else:
            result.append('})')
        
        return ''.join(result)
    
    #__setattr__ -> same
    
    @has_docs
    def __setitem__(self, key, value):
        """Adds the given `key` - `value` pair to the weak item dictionary."""
        key_reference = WeakReferer(key, self._key_callback)
        value_reference = KeyedReferer(value, self._value_callback, key_reference)
        dict.__setitem__(self, key_reference, value_reference)
    
    # __sizeof__ -> same
    
    __str__ = __repr__
    
    # __subclasshook__ -> same
    
    @has_docs
    def clear(self):
        """
        Clears the weak item dictionary.
        """
        dict.clear(self)
        self._pending_removals = None
    
    
    @has_docs
    def copy(self):
        """
        Copies the weak item dictionary.
        
        Returns
        -------
        new : ``WeakItemDictionary``
        """
        new = dict.__new__(type(self))
        new._pending_removals = None
        key_callback = _WeakKeyDictionaryCallback(new)
        new._key_callback = key_callback
        value_callback = _WeakItemDictionaryValueCallback(new)
        new._value_callback = value_callback
        new._iterating = 0
        
        for key_reference, value_reference in dict.items(self):
            key = key_reference()
            if key is None:
                add_to_pending_removals(self, key_reference)
                continue
            
            value = value_reference()
            if value is None:
                add_to_pending_removals(self, key_reference)
                continue
            
            key_reference = WeakReferer(key, self._key_callback)
            value_reference = KeyedReferer(value, self._value_callback, key_reference)
            
            dict.__setitem__(new, key_reference, value_reference)
            continue
        
        self._commit_removals()
        
        return new
    
    
    @has_docs
    def get(self, key, default=None):
        """
        Gets the value of the weak item dictionary which matches the given key.
        
        Parameters
        ----------
        key : `Any`
            A key to match.
        default : `None`, `Any` = `None`, Optional
            Default value to return if the given `key` could not be matched.
        
        Returns
        -------
        value : `Any`, `default`
            The key's matched value. If no value was matched returns the `default` value.
        """
        try:
            key_reference = WeakReferer(key)
        except TypeError:
            pass
        else:
            value_reference = dict.get(self, key_reference, default)
            if value_reference is default:
                return default
            
            value = value_reference()
            if (value is not None):
                return value
            
            if self._iterating:
                add_to_pending_removals(self, value_reference.key)
            else:
                dict.__delitem__(self, key)
        
        return default
    
    
    @has_docs
    def items(self):
        """
        Returns item iterator for the weak item dictionary.
        
        Returns
        -------
        item_iterator : ``_WeakItemDictionaryItemIterator``
        """
        return _WeakItemDictionaryItemIterator(self)
    
    
    @has_docs
    def keys(self):
        """
        Returns key iterator for the weak item dictionary.
        
        Returns
        -------
        key_iterator : ``_WeakKeyDictionaryKeyIterator``
        """
        return _WeakKeyDictionaryKeyIterator(self)
    
    
    @has_docs
    def pop(self, key, default=...):
        """
        Pops the value of the weak item dictionary which matches the given key.
        
        Parameters
        ----------
        key : `Any`
            A key to match.
        default : `Any`, Optional
            Default value to return if the given `key` could not be matched.
        
        Returns
        -------
        value : `Any`, `default`
            The key's matched value. If no value was matched and `default` value is given, then returns that.
        
        Raises
        ------
        KeyError
            If `key` could not be matched and `default` value is was not given either.
        """
        # Use goto
        while True:
            try:
                key_reference = WeakReferer(key)
            except TypeError:
                break
            
            value_reference = dict.pop(self, key_reference, ...)
            if (value_reference is ...):
                break
                
            value = value_reference()
            if (value is None):
                break
            
            return value
        
        
        if default is ...:
            raise KeyError(key)
        
        return default
    
    
    @has_docs
    def popitem(self):
        """
        Pops an item of the weak item dictionary.
        
        Returns
        -------
        item : `tuple` (`Any`, `Any`)
        
        Raises
        ------
        KeyError
            If the weak item dictionary is empty.
        """
        while dict.__len__(self):
            key_reference, value_reference = dict.popitem(self)
            key = key_reference()
            if key is None:
                continue
            
            value = value_reference()
            if value is None:
                continue
            
            return key, value
        
        raise KeyError('popitem(): dictionary is empty.')
    
    
    @has_docs
    def setdefault(self, key, default):
        """
        Returns the value for the given `key`.
        
        If the `key` is not present in the weak item dictionary, then set's the given `default` value as it.
        
        Parameters
        ----------
        key : `Any`
            The key to match.
        default : `Any`
            Default value to set and return if `key` is not present in the weak item dictionary.
        
        Returns
        -------
        value : `default`, `Any`
            The matched value, or `default` if none.
        """
        value_reference = dict.get(self, WeakReferer(key), ...)
        if (value_reference is not ...):
            value = value_reference()
            if (value is not None):
                return value
        
        self[key] = default
        return default
    
    
    @has_docs
    def update(self, iterable):
        """
        Updates the weak item dictionary with the given iterable's elements.
        
        Parameters
        ----------
        iterable : `iterable`
            Iterable to extend the weak item dictionary with.
            
            Can be given as an object, which:
            - supports `.items` iterator.
            - supports `.keys` and `.__getitem__`.
            - is `iterable` and each iteration returns a sequence with 2 elements.
        
        Raises
        ------
        TypeError
            The given `iterable` is not `iterable`.
        ValueError
            The the given `iterable` sot supports neither `.items` or (`.keys` and `.__getitem__`) and one of it's
            element's length is not `2`.
        """
        iterable_type = iterable.__class__
        if hasattr(iterable_type, 'items'):
            for key, value in iterable.items():
                self[key] = value
            return
        
        if hasattr(iterable_type, 'keys') and hasattr(iterable_type, '__getitem__'):
            for key in iterable.keys():
                value = iterable[key]
                self[key] = value
            return
        
        if hasattr(iterable_type, '__iter__'):
            for key, value in _dict_update_iterable_iterator(iterable):
                self[key] = value
            
            return
        
        raise TypeError(
            f'`iterable` can be `iterable`, got {iterable_type.__name__}; {iterable_type!r}'
        )
    
    
    @has_docs
    def values(self):
        """
        Returns value iterator for the weak item dictionary.
        
        Returns
        -------
        value_iterator : ``_WeakItemDictionaryValueIterator``
        """
        return _WeakItemDictionaryValueIterator(self)



from .ansi import *
from .constants import *
from .default import *
from .formatter_context import *
from .formatter_detail import *
from .formatter_node import *
from .matching import *
from .parser_context import *
from .token import *
from .token_types import *
from .utils import *
from .word_pattern import *

from . import token_types as HIGHLIGHT_TOKEN_TYPES


__all__ = (
    'HIGHLIGHT_TOKEN_TYPES',
    
    *ansi.__all__,
    *constants.__all__,
    *default.__all__,
    *formatter_context.__all__,
    *formatter_detail.__all__,
    *formatter_node.__all__,
    *matching.__all__,
    *parser_context.__all__,
    *token.__all__,
    *token_types.__all__,
    *utils.__all__,
    *word_pattern.__all__,
)



__all__ = ('Server', )

from ..traps import Future, skip_ready_cycle


class Server:
    """
    Server returned by ``EventThread.create_server``.
    
    Attributes
    ----------
    active_count : `int`
        The amount of active connections bound to the server.
    backlog : `int`
        The maximum number of queued connections passed to `listen()` (defaults to 100).
    close_waiters : `None`, `list` of ``Future``
        Futures, which are waiting for the server to close. If the server is already closed, set as `None`.
    loop : ``EventThread``
        The event loop to what the server is bound to.
    protocol_factory : `callable`
        Factory function for creating a protocols.
    serving : `bool`
        Whether the server is serving.
    sockets : `None`, `list` of `socket.socket`
        The sockets served by the server. If the server is closed, then i set as `None`.
    ssl_context : `None`, `ssl.SSLContext`
        If ssl is enabled for the connections, then set as `ssl.SSLContext`.
    """
    __slots__ = (
        'active_count', 'backlog', 'close_waiters', 'loop', 'protocol_factory', 'serving', 'sockets', 'ssl_context'
    )
    
    def __init__(self, loop, sockets, protocol_factory, ssl_context, backlog):
        """
        Creates a new server with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the server will be bound to.
        sockets : `list` of `socket.socket`
            The sockets to serve by the server.
        protocol_factory : `callable`
            Factory function for creating a protocols.
        ssl_context : `None`, `ssl.SSLContext`
            To enable ssl for the connections, give it as  `ssl.SSLContext`.
        backlog : `int`
            The maximum number of queued connections passed to `listen()` (defaults to 100).
        """
        self.loop = loop
        self.sockets = sockets
        self.active_count = 0
        self.close_waiters = []
        self.protocol_factory = protocol_factory
        self.backlog = backlog
        self.ssl_context = ssl_context
        self.serving = False
    
    
    def __repr__(self):
        """Returns the server's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        if self.serving:
            repr_parts.append(' serving')
            
        repr_parts.append(' sockets=')
        repr_parts.append(repr(self.sockets))
        
        repr_parts.append(', protocol_factory=')
        repr_parts.append(repr(self.protocol_factory))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _attach(self):
        """
        Adds `1` to the server active counter.
        """
        self.active_count += 1
    
    
    def _detach(self):
        """
        Removes `1` from the server's active counter. If there no more active sockets of the server, then closes it.
        """
        active_count = self.active_count - 1
        self.active_count = active_count
        if active_count:
            return
        
        if (self.sockets is None):
            self._wake_up_close_waiters()


    def _wake_up_close_waiters(self):
        """
        Wakes up the server's close waiters.
        """
        close_waiters = self.close_waiters
        if close_waiters is None:
            return
        
        self.close_waiters = None
        for close_waiter in close_waiters:
            close_waiter.set_result(None)
    
    
    def close(self):
        """
        Closes the server by stopping serving it's sockets and waking up it's close waiters.
        """
        sockets = self.sockets
        if sockets is None:
            return
        
        self.sockets = None
        
        loop = self.loop
        for socket in sockets:
            loop._stop_serving(socket)
        
        self.serving = False
        
        if self.active_count == 0:
            self._wake_up_close_waiters()
    
    
    async def start(self):
        """
        Starts the server by starting serving it's sockets.
        
        This method is a coroutine.
        """
        if self.serving:
            return
        
        self.serving = True
        
        protocol_factory = self.protocol_factory
        ssl_context = self.ssl_context
        backlog = self.backlog
        loop = self.loop
        
        for socket in self.sockets:
            socket.listen(backlog)
            loop._start_serving(protocol_factory, socket, ssl_context, self, backlog)
        
        # Skip one event loop cycle, so all the callbacks added up ^ will run before returning.
        await skip_ready_cycle()
    

    async def wait_closed(self):
        """
        Blocks the task, till the sever is closes.
        
        This method is a coroutine.
        """
        if self.sockets is None:
            return
        
        close_waiters = self.close_waiters
        if close_waiters is None:
            return
        
        close_waiter = Future(self.loop)
        close_waiters.append(close_waiter)
        await close_waiter



__all__ = ()


class FormatterNode:
    """
    Represents a formatter node for a specified token type.
    
    Attributes
    ----------
    detail : `None`, ``FormatterDetail``
        Details of the node for formatting.
    direct : `bool`
        Whether ``.type`` is set directly or is inherited.
    id : `int`
        The node's identifier.
    nodes : `None`, `dict` of (`int`, ``FormatterNode``) items
        Sub-nodes branching out from the source one.
    parent : `None`, ``FormatterNode``
        The parent node.
    """
    __slots__ = ('detail', 'direct', 'id', 'nodes', 'parent')
    
    def __new__(cls, formatter, id_):
        """
        Creates a new ``TokenClassNode`` with the given identifier.
        
        Parameters
        ----------
        formatter : ``HighlightFormatterContext``
            The owner formatter.
        id_ : `int`
            The identifier of the token.
        """
        self = object.__new__(cls)
        self.detail = None
        self.direct = False
        self.id = id_
        self.nodes = None
        self.parent = None
        
        formatter.formatter_nodes[id_] = self
        
        return self
    
    
    def __repr__(self):
        """Returns the token type node's representation."""
        result = ['<', self.__class__.__name__, ' id=', repr(self.id)]
        
        if self.direct:
            result.append(', detail=')
            result.append(repr(self.detail))
        
        nodes = self.nodes
        if (nodes is not None):
            result.append(', nodes=')
            result.append(repr(nodes))
        
        result.append('>')
        
        return ''.join(result)
    
    
    def add_node(self, node):
        """
        Adds a sub node.
        
        Parameters
        ----------
        node : ``FormatterNode``
        """
        node.parent = self
        
        nodes = self.nodes
        if nodes is None:
            self.nodes = nodes = {}
        
        nodes[node.id] = node
    
    
    def _set_detail(self, detail):
        """
        Sets the node's html class. This is an internal method called by ``.set_detail`` or by itself recursively
        to actually modify the attributes.
        
        This method shall be called only from a parent node.
        
        Parameters
        ----------
        detail : `None`, ``FormatterDetail``
            Details for formatting, set or `None` to remove.
        """
        self.detail = detail
        
        nodes = self.nodes
        if (nodes is not None):
            for node in nodes.values():
                if not node.direct:
                    node._set_detail(detail)
    
    
    def set_detail(self, detail, *, direct=True):
        """
        Sets html class to the node.
        
        Parameters
        ----------
        detail : `None`, ``FormatterDetail``
            Details for formatting, set or `None` to remove.
        direct : `bool` = `True`, Optional (Keyword only)
            Whether we are marking it as a direct set.
        """
        if detail is None:
            if self.direct:
                self.direct = False
                parent = self.parent
                if (parent is not None):
                    detail = parent.detail
                    # Only continue if the html classes are different
                    if (self.detail != detail):
                        self._set_detail(detail)
        
        else:
            self.direct = direct
            self._set_detail(detail)



__all__ = ()

import socket as module_socket
import threading
from threading import Event as SyncEvent, Thread, current_thread

from ...utils import ignore_frame


ignore_frame(threading.__spec__.origin, '_bootstrap', 'self._bootstrap_inner()', )
ignore_frame(threading.__spec__.origin, '_bootstrap_inner', 'self.run()', )


class EventThreadContextManager:
    """
    Context manager of an ``EventThread``, which wraps it's runner. when the runner is started up, set it's ``waiter.``
    allowing the starter thread to continue.
    
    Attributes
    ----------
    thread : `None`, ``EventThread``
        The wrapped event loop.
    thread_waiter : `None`, `threading.Event`
       Threading event, what is set, when the thread is started up. Set as `None` after set.
    """
    __slots__ = ('thread', 'thread_waiter',)
    
    
    def __new__(cls, thread):
        """
        Creates a new event thread context.
        
        Parameters
        ----------
        thread : ``EventThread``
            The event thread to wrap.
        
        Returns
        -------
        self : ``EventThreadContextManager``
            The created instance
        thread_waiter : `threading.Event`
            Threading event, what is set, when the thread is started up.
        """
        self = object.__new__(cls)
        self.thread = thread
        self.thread_waiter = SyncEvent()
        return self
    
    
    def __enter__(self):
        """
        Called, when the respective event loop's runner started up.
        
        Enters the event loop runner setting it's waiter and finishes the loop's initialization.
        
        Raises
        ------
        RuntimeError
            - If ``EventThreadContextManager.__enter__`` was called a second time.
            - If called from a different thread as is bound to.
            - If the ``EventThread`` is already running.
            - If the ``EventThread`` is already stopped.
        """
        thread_waiter = self.thread_waiter
        if thread_waiter is None:
            raise RuntimeError(
                f'`{self.__class__.__name__}.__enter__` called with thread waiter lock set.'
            )
        
        try:
            thread = self.thread
            if (thread is not current_thread()):
                raise RuntimeError(
                    f'`{thread!r}.run` called from an other thread: {current_thread()!r}'
                )
            
            if (thread.running):
                raise RuntimeError(
                    f'`{thread!r}.run` called when the thread is already running.'
                )
            
            if (thread._is_stopped):
                raise RuntimeError(
                    f'`{thread!r}.run` called when the thread is already stopped.'
                )
            
            thread.running = True
            
            self_read_socket, self_write_socket = module_socket.socketpair()
            self_read_socket.setblocking(False)
            self_write_socket.setblocking(False)
            thread._self_read_socket = self_read_socket
            thread._self_write_socket = self_write_socket
            thread.add_reader(self_read_socket.fileno(), thread.empty_self_socket)
        finally:
            thread_waiter.set()
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        When the event loop's runner stops, it's context closes it.
        """
        thread = self.thread
        self.thread = None
        
        thread.running = False
        thread.remove_reader(thread._self_read_socket.fileno())
        try:
            thread._self_read_socket.close()
        except OSError as err:
            if err.errno != 9:
                raise
        finally:
            thread._self_read_socket = None
        
        try:
            thread._self_write_socket.close()
        except OSError as err:
            if err.errno != 9:
                raise
        finally:
            thread._self_write_socket = None
        
        thread._ready.clear()
        thread._scheduled.clear()
        
        thread.cancel_executors()
        
        selector = thread.selector
        if (selector is not None):
            thread.selector = None
            selector.close()
        
        return False



class EventThreadType(type):
    """
    Type of even thread, which manages their instances creation.
    """
    def __call__(cls, daemon=False, name=None, start_later=True, **kwargs):
        """
        Creates a new ``EventThread`` with the given parameters.
        
        Parameters
        ----------
        daemon : `bool` = `False`, Optional
            Whether the created thread should be daemon. Defaults to `False`.
        name : `None`, `str` = `None`, Optional
            The created thread's name. Defaults to `None`
        start_later : `bool` = `True`, Optional
            Whether the event loop should be started only later. Defaults to `True`.
        **kwargs : keyword parameters
            Additional event thread specific parameters.
        
        Returns
        -------
        obj : ``EventThread``
            The created event loop.
        """
        obj = Thread.__new__(cls)
        cls.__init__(obj, **kwargs)
        Thread.__init__(obj, daemon=daemon, name=name)
        
        obj.context = EventThreadContextManager(obj)
        
        if not start_later:
            obj._do_start()
        
        return obj



__all__ = ('Task',)

import reprlib, sys
from threading import current_thread
from types import AsyncGeneratorType as CoroutineGeneratorType, CoroutineType, GeneratorType

from ...utils import alchemy_incendiary, copy_docs, ignore_frame, include, render_frames_into
from ...utils.trace import format_callback, format_coroutine, render_exception_into

from ..exceptions import CancelledError, InvalidStateError

from .future import (
    FUTURE_STATE_FINISHED, FUTURE_STATE_PENDING, FUTURE_STATE_RETRIEVED, Future, get_exception_short_representation,
    get_future_state_name
)


ignore_frame(__spec__.origin, 'get_result', 'raise exception')
ignore_frame(__spec__.origin, '__iter__', 'yield self')
ignore_frame(__spec__.origin, '_step', 'result = coroutine.throw(exception)')
ignore_frame(__spec__.origin, '_step', 'result = coroutine.send(None)')
ignore_frame(__spec__.origin, '_wake_up', 'future.get_result()')
ignore_frame(__spec__.origin, '__call__', 'future.get_result()')

EventThread = include('EventThread')


class Task(Future):
    """
    A Future-like object that runs a Python coroutine.
    
    Tasks are used to run coroutines in event loops. If a coroutine awaits on a ``Future`, the ``Task`` suspends the
    execution of the coroutine and waits for the completion of the ``Future``. When the ``Future`` is done, the
    execution of the wrapped coroutine resumes.
    
    Attributes
    ----------
    _blocking : `bool`
        Whether the task is already being awaited, so it blocks the respective coroutine.
    _callbacks : `list` of `callable`
        The callbacks of the task, which are queued up on the respective event loop to be called, when the task is
        finished. These callback should accept `1` parameter, the task itself.
        
        Note, if the task is already done, then the newly added callbacks are queued up instantly on the respective
        event loop to be called.
    
    _exception : `None`, `BaseException`
        The exception raised by task's internal coroutine. Defaults to `None`.
    _loop : ``EventThread``
        The loop to what the created task is bound.
    _result : `None`, `Any`
        The result of the task. Defaults to `None`.
    _state : `str`
        The state of the task.
        
        Can be set as one of the following:
        
        +---------------------------+-----------+
        | Respective name           | Value     |
        +===========================+===========+
        | FUTURE_STATE_PENDING      | `0`       |
        +---------------------------+-----------+
        | FUTURE_STATE_CANCELLED    | `1`       |
        +---------------------------+-----------+
        | FUTURE_STATE_FINISHED     | `2`       |
        +---------------------------+-----------+
        | FUTURE_STATE_RETRIEVED    | `3`       |
        +---------------------------+-----------+
        
        Note, that states are checked by memory address and not by equality. Also ``FUTURE_STATE_RETRIEVED`` is used
        only if `__debug__` is set as `True`.
    
    _coroutine : `CoroutineType`, `GeneratorType`
        The wrapped coroutine.
    _must_cancel : `bool`
        Whether the task is cancelled, and at it's next step a ``CancelledError`` would be raised into it's coroutine.
    _waited_future : `None`, ``Future``
        The future on what's result the future is waiting right now.
    """
    __slots__ = ('_coroutine', '_must_cancel', '_waited_future')
    
    def __new__(cls, coroutine, loop):
        """
        Creates a new ``Task`` object running the given coroutine on the given event loop.
        
        Parameters
        ----------
        coroutine : `CoroutineType`, `GeneratorType`
            The coroutine, what the task will on the respective event loop.
        loop : ``EventThread``
            The event loop on what the coroutine will run.
        """
        self = object.__new__(cls)
        self._loop = loop
        self._state = FUTURE_STATE_PENDING

        self._result = None
        self._exception = None
        
        self._callbacks = []
        self._blocking = False

        self._must_cancel = False
        self._waited_future = None
        self._coroutine = coroutine
        
        loop.call_soon(self._step)

        return self
    
    
    def get_stack(self, limit=-1):
        """
        Return the list of stack frames for the task. If the task is already done, returns an empty list.
        
        Parameters
        ----------
        limit : `int` = `-1`, Optional
            The maximal amount of stacks to fetch. By giving it as negative integer, there will be no stack limit
            to fetch back.
        
        Returns
        -------
        frames : `list` of `frame`
            The stack frames of the task.
        """
        frames = []
        
        coroutine = self._coroutine
        if isinstance(coroutine, GeneratorType):
            frame = coroutine.gi_frame
        elif isinstance(coroutine, CoroutineType):
            frame = coroutine.cr_frame
        elif isinstance(coroutine, CoroutineGeneratorType):
            frame = coroutine.ag_frame
        else:
            frame = None
        
        if frame is None:
            return frames
        
        while limit:
            limit -= 1
            
            if frame in frames:
                frames.append(frame)
                frames.append(None)
                return frames
            
            frames.append(frame)
            
            if isinstance(coroutine, GeneratorType):
                coroutine = coroutine.gi_yieldfrom
            elif isinstance(coroutine, CoroutineType):
                coroutine = coroutine.cr_await
            elif isinstance(coroutine, CoroutineGeneratorType):
                coroutine = coroutine.ag_await
            else:
                coroutine = None
            
            if coroutine is not None:
                if isinstance(coroutine, GeneratorType):
                    frame = coroutine.gi_frame
                elif isinstance(coroutine, CoroutineType):
                    frame = coroutine.cr_frame
                elif isinstance(coroutine, CoroutineGeneratorType):
                    frame = coroutine.ag_frame
                else:
                    frame = None
                
                if frame is not None:
                    continue
            
            self = self._waited_future
            if self is None:
                break
            
            del frames[-1]
            if not isinstance(self, Task):
                break
            
            coroutine = self._coroutine
            
            if isinstance(coroutine, GeneratorType):
                frame = coroutine.gi_frame
            elif isinstance(coroutine, CoroutineType):
                frame = coroutine.cr_frame
            elif isinstance(coroutine, CoroutineGeneratorType):
                frame = coroutine.ag_frame
            else:
                frame = None
                
            if frame is None:
                break
        
        return frames
    
    
    def __repr__(self):
        """Returns the task's representation."""
        repr_parts = ['<', self.__class__.__name__, ' ']
        
        state = self._state
        repr_parts.append(get_future_state_name(state))
        
        if self._must_cancel:
            repr_parts.append(' (cancelling)')
        
        repr_parts.append(' coroutine=')
        repr_parts.append(format_coroutine(self._coroutine))
        
        waited_future = self._waited_future
        if waited_future is not None:
            repr_parts.append(', waits for=')
            if type(waited_future) is type(self):
                repr_parts.append(waited_future.qualname)
            else:
                repr_parts.append(repr(waited_future))
        
        if (not self._must_cancel) and (state >= FUTURE_STATE_FINISHED):
            exception = self._exception
            if exception is None:
                repr_parts.append(', result=')
                repr_parts.append(reprlib.repr(self._result))
            else:
                repr_parts.append(', exception=')
                repr_parts.append(get_exception_short_representation(exception))
        
        callbacks = self._callbacks
        limit = len(callbacks)
        if limit:
            repr_parts.append(', callbacks=[')
            index = 0
            while True:
                callback = callbacks[index]
                repr_parts.append(format_callback(callback))
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def print_stack(self, limit=-1, file=None):
        """
        Prints the stack or traceback of the task.
        
        Parameters
        ----------
        limit : `int` = `-1`, Optional
            The maximal amount of stacks to print. By giving it as negative integer, there will be no stack limit
            to print out.
        file : `None`, `I/O stream` = `None`, Optional
            The file to print the stack to. Defaults to `sys.stderr`.
        
        Notes
        -----
        If the task is finished with an exception, then `limit` is ignored when printing traceback.
        """
        local_thread = current_thread()
        if isinstance(local_thread, EventThread):
            return local_thread.run_in_executor(alchemy_incendiary(self._print_stack, (self, limit, file),))
        else:
            self._print_stack(self, limit, file)
    
    
    @staticmethod
    def _print_stack(self, limit, file):
        """
        Prints the stack or traceback of the task to the given `file`.
        
        Parameters
        ----------
        limit : `int`
            The maximal amount of stacks to print. By giving it as negative integer, there will be no stack limit
            to print out,
        file : `None`, `I/O stream`
            The file to print the stack to. Defaults to `sys.stderr`.
        
        Notes
        -----
        This function calls blocking operations and should not run inside of an event loop.
        
        If the task is finished with an exception, then `limit` is ignored when printing traceback.
        """
        if file is None:
            file = sys.stdout
        
        exception = self._exception
        
        if exception is None:
            frames = self.get_stack(limit)
            if frames:
                recursive = (frames[-1] is None)
                if recursive:
                    del frames[-1]
                
                extracted = ['Stack for ', repr(self), ' (most recent call last):\n']
                extracted = render_frames_into(frames, extend=extracted)
                if recursive:
                    extracted.append(
                        'Last frame is a repeat from a frame above. Rest of the recursive part is not rendered.'
                    )
            else:
                extracted = ['No stack for ', repr(self), '\n']
        else:
            extracted = ['Traceback for ', repr(self), ' (most recent call last):\n']
            extracted = render_exception_into(exception, extend=extracted)
        
        file.write(''.join(extracted))
    
    if __debug__:
        @copy_docs(Future.cancel)
        def cancel(self):
            state = self._state
            if state != FUTURE_STATE_PENDING:
                if state == FUTURE_STATE_FINISHED:
                    self._state = FUTURE_STATE_RETRIEVED
                
                return 0
            
            waited_future = self._waited_future
            if (waited_future is None) or (not waited_future.cancel()):
                self._must_cancel = True
            
            return 1
        
    else:
        @copy_docs(Future.cancel)
        def cancel(self):
            if self._state != FUTURE_STATE_PENDING:
                return 0
            
            waited_future = self._waited_future
            if (waited_future is None) or (not waited_future.cancel()):
                self._must_cancel = True
            
            return 1
    
    
    @property
    def name(self):
        """
        Returns the task's wrapped coroutine's name.
        
        Returns
        -------
        name : `str`
        """
        coroutine = self._coroutine
        try:
            return coroutine.__name__
        except AttributeError:
            return coroutine.__class__.__name__
    
    
    @property
    def qualname(self):
        """
        Returns the task's wrapped coroutine's qualname.
        
        Returns
        -------
        qualname : `str`
        """
        coroutine = self._coroutine
        try:
            return coroutine.__qualname__
        except AttributeError:
            return coroutine.__class__.__qualname__
        
        
    def set_result(self, result):
        """
        Tasks do not support `.set_result` operation.
        
        Parameters
        ----------
        result : `Any`
            The object to set as result.
        
        Raises
        ------
        RuntimeError
            Tasks do not support `.set_result` operation.
        """
        raise RuntimeError(
            f'`{self.__class__.__name__}` does not support `.set_result` operation.'
        )
    
    
    def set_result_if_pending(self, result):
        """
        Tasks do not support `.set_result_if_pending` operation.
        
        Parameters
        ----------
        result : `Any`
            The object to set as result.
        
        Raises
        ------
        RuntimeError
            Tasks do not support `.set_result_if_pending` operation.
        """
        raise RuntimeError(
            f'`{self.__class__.__name__}` does not support `.set_result_if_pending` operation.'
        )
    
    
    # We will not send an exception to a task, but we will cancel it.
    # The exception will show up as ``._exception`` tho.
    # We also wont change the state of the Task, it will be changed, when the next `._step` is done with the
    # #cancelling.
    @copy_docs(Future.set_exception)
    def set_exception(self, exception):
        if self._state != FUTURE_STATE_PENDING:
            raise InvalidStateError(self, 'set_exception')
        
        if (self._waited_future is None) or self._waited_future.cancel():
            self._must_cancel = True
        
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
            )
        
        self._exception = exception
    
    
    @copy_docs(Future.set_exception_if_pending)
    def set_exception_if_pending(self, exception):
        if self._state != FUTURE_STATE_PENDING:
            return 0
        
        if (self._waited_future is None) or self._waited_future.cancel():
            self._must_cancel = True
        
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
            )
        
        self._exception = exception
        return 1
    
    
    def clear(self):
        """
        Tasks do not support `.clear` operation.
        
        Raises
        ------
        RuntimeError
            Tasks do not support `.clear` operation.
        """
        raise RuntimeError(
            f'`{self.__class__.__name__}` does not support `.clear` operation.'
        )
    
    
    def _step(self, exception=None):
        """
        Does a step, by giving control to the wrapped coroutine by the task.
        
        If `exception` is given, then that exception will be raised to the internal coroutine, exception if the task
        is already cancelled, because, then the exception to raise will be decided by ``._must_exception``.
        
        Parameters
        ----------
        exception : `None`, `BaseException`
            Exception to raise into the wrapped coroutine.
        
        Raises
        ------
        InvalidStateError
            If the task is already done.
        """
        if self._state != FUTURE_STATE_PENDING:
            raise InvalidStateError(
                self,
                '_step',
                message = f'`{self.__class__.__name__}._step` already done of {self!r}; exception={exception!r}.',
            )
        
        if self._must_cancel:
            exception = self._must_exception(exception)
        
        coroutine = self._coroutine
        self._waited_future = None
        
        self._loop.current_task = self
        
        # call either coroutine.throw(err) or coroutine.send(None).
        try:
            if exception is None:
                result = coroutine.send(None)
            else:
                result = coroutine.throw(exception)
        except StopIteration as exception:
            if self._must_cancel:
                # the task is cancelled meanwhile
                self._must_cancel = False
                Future.set_exception(self, CancelledError())
            else:
                Future.set_result(self, exception.value)
        except CancelledError:
            Future.cancel(self)
        except BaseException as exception:
            Future.set_exception(self, exception)
        else:
            if isinstance(result, Future) and result._blocking:
                result._blocking = False
                result.add_done_callback(self._wake_up)
                self._waited_future = result
                if self._must_cancel:
                    if result.cancel():
                        self._must_cancel = False
            else:
                self._loop.call_soon(self._step)
        finally:
            self._loop.current_task = None
            self = None # Need to set `self` as `None`. Else `self` might never get garbage collected.
    
    
    def _wake_up(self, future):
        """
        Callback used by ``._step`` when the wrapped coroutine waits on a future to be marked as done.
        
        Parameters
        ----------
        future : ``Future``
            The future for what's completion the task is waiting for.
        """
        try:
            future.get_result()
        except BaseException as err:
            self._step(err)
        else:
            self._step()
        
        self = None # set self as `None`, so when exception occurs, self can be garbage collected.
    
    
    def _must_exception(self, exception):
        """
        Returns the exception, what should be raised into the tasks' wrapped coroutine.
        
        Parameters
        ----------
        exception : `None`, `BaseException`
            The exception, what would be preferable raised into the task.
        
        Returns
        -------
        exception : `BaseException`
            If task has already `˙._exception`˙ set, returns that. If `exception` is given as `None`, or as non
            ``CancelledError``, will create a new ``CancelledError`` and return that.
        """
        self_exception = self._exception
        if self_exception is None:
            if (exception is None) or (not isinstance(exception, CancelledError)):
                exception = CancelledError()
        else:
            exception = self_exception
        
        self._must_cancel = False
        return exception



__all__ = ('FutureSyncWrapper', )

import reprlib, sys, warnings
from threading import Event as SyncEvent, Lock as SyncLock
from types import MethodType

from ...utils import copy_docs, export, ignore_frame, include, set_docs

from ..exceptions import CancelledError, InvalidStateError

from .future import (
    FUTURE_STATE_CANCELLED, FUTURE_STATE_FINISHED, FUTURE_STATE_PENDING, FUTURE_STATE_RETRIEVED, Future,
    get_future_state_name, get_exception_short_representation
)


ignore_frame(__spec__.origin, 'get_result', 'raise exception')
ignore_frame(__spec__.origin, 'wait', 'return self.get_result()')

write_exception_maybe_async = include('write_exception_maybe_async')


@export
class FutureSyncWrapper:
    """
    Sync wrapper for ``Future``-s enabling them to be waited from a sync threads.
    
    Attributes
    ----------
    _exception : `None`, `BaseException`
        The exception set to the future as it's result. Defaults to `None`.
    _future : `None`, ``Future``
        The waited future. If the future's state == modified by the sync wrapper, then ``._future`` is set as `None`,
        to not retrieve the result again.
    _lock : `threading.Lock`
        Threading lock to disable concurrent access to the future.
    _loop : ``EventLoop``
        The event loop of the future sync wrapper to which it is bound to.
    _result : `None`, `Any`
        The result of the future. Defaults to `None`.
    _state : `str`
        The state of the future.
        
        Can be set as one of the following:
        
        +---------------------------+-----------+
        | Respective name           | Value     |
        +===========================+===========+
        | FUTURE_STATE_PENDING      | `0`       |
        +---------------------------+-----------+
        | FUTURE_STATE_CANCELLED    | `1`       |
        +---------------------------+-----------+
        | FUTURE_STATE_FINISHED     | `2`       |
        +---------------------------+-----------+
        | FUTURE_STATE_RETRIEVED    | `3`       |
        +---------------------------+-----------+
        
        Note, that states are checked by memory address and not by equality. Also ``FUTURE_STATE_RETRIEVED`` is
        used only if `__debug__` is set as `True`.
    _waiter : `threading.Event`
        An event, what is set, when the waited future is done.
    """
    __slots__ = ('_exception', '_future', '_lock', '_loop', '_result', '_state', '_waiter')
    
    def __new__(cls, future):
        """
        Creates a new ``FutureSyncWrapper`` wrapping the given `future`
        
        Parameters
        ----------
        future : ``Future``
            The future on what we would want to wait from a sync thread.
        """
        loop = future._loop
        
        self = object.__new__(cls)
        self._future = future
        self._lock = SyncLock()
        self._loop = loop
        self._waiter = SyncEvent()
        self._state = FUTURE_STATE_PENDING
        self._result = None
        self._exception = None
        
        loop.call_soon(future.add_done_callback, self._done_callback)
        loop.wake_up()
        
        return self
    
    
    def __call__(self, future):
        """
        By calling a ``FutureSyncWrapper`` you can make it bound to an other future.
        
        Parameters
        ----------
        future : ``Future``
            The future on what you would want to wait from a sync thread.
        
        Returns
        -------
        self : ``FutureSyncWrapper``
        """
        with self._lock:
            old_future = self._future

            if old_future is not None:
                old_loop = self._loop
                old_loop.call_soon(self._remove_callback, old_future)
                old_loop.wake_up()
            
            new_loop = future._loop
            
            self._future = future
            self._state = FUTURE_STATE_PENDING
            self._result = None
            self._exception = None
            self._waiter.clear()
            self._loop = new_loop
            
            new_loop.call_soon(future.add_done_callback, self._done_callback)
            new_loop.wake_up()

        return self
    
    
    def __repr__(self):
        """Returns the future sync wrapper's representation."""
        repr_parts = ['<', self.__class__.__name__, ' ', ]
        
        state = self._state
        repr_parts.append(get_future_state_name(state))
        
        if state >= FUTURE_STATE_FINISHED:
            
            exception = self._exception
            if exception is None:
                repr_parts.append(' result=')
                repr_parts.append(reprlib.repr(self._result))
            else:
                repr_parts.append(' exception=')
                repr_parts.append(get_exception_short_representation(exception))
        
        future = self._future
        if future is not None:
            # we do not want to repr it, keep it thread safe
            repr_parts.append(' future=')
            repr_parts.append(future.__class__.__name__)
            repr_parts.append('(...)')
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    if __debug__:
        @copy_docs(Future.cancel)
        def cancel(self):
            state = self._state
            if state != FUTURE_STATE_PENDING:
                if state == FUTURE_STATE_FINISHED:
                    self._state = FUTURE_STATE_RETRIEVED
                
                return 0
            
            future = self._future
            if future is None:
                self._state = FUTURE_STATE_CANCELLED
                self._waiter.set()
                return 1
            
            loop = self._loop
            loop.call_soon(future.cancel)
            loop.wake_up()
            return 1
        
    else:
        @copy_docs(Future.cancel)
        def cancel(self):
            if self._state != FUTURE_STATE_PENDING:
                return 0
            
            future = self._future
            if future is None:
                self._state = FUTURE_STATE_CANCELLED
                self._waiter.set()
                return 1
            
            loop = self._loop
            loop.call_soon(future.cancel)
            loop.wake_up()
            return 1
    
    
    @copy_docs(Future.is_cancelled)
    def is_cancelled(self):
        return (self._state == FUTURE_STATE_CANCELLED)
    
    
    def cancelled(self):
        warnings.warn(
            f'`{self.__class__.__name__}.cancelled` is deprecated.',
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_cancelled()
    
    
    @copy_docs(Future.is_done)
    def is_done(self):
        return (self._state != FUTURE_STATE_PENDING)
    
    
    def done(self):
        warnings.warn(
            f'`{self.__class__.__name__}.done` is deprecated.',
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_done()
    
    
    @copy_docs(Future.is_pending)
    def is_pending(self):
        return (self._state == FUTURE_STATE_PENDING)
    
    
    def pending(self):
        warnings.warn(
            f'`{self.__class__.__name__}.pending` is deprecated.',
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_pending()
    
    
    if __debug__:
        @copy_docs(Future.result)
        def get_result(self):
            with self._lock:
                state = self._state
                
                if state == FUTURE_STATE_FINISHED:
                    self._state = FUTURE_STATE_RETRIEVED
                    exception = self._exception
                    if exception is None:
                        return self._result
                    raise exception
                
                if state == FUTURE_STATE_RETRIEVED:
                    exception = self._exception
                    if exception is None:
                        return self._result
                    raise exception
                
                if state == FUTURE_STATE_CANCELLED:
                    raise CancelledError
            
            # still pending
            raise InvalidStateError(self, 'result')
    
    else:
        @copy_docs(Future.result)
        def get_result(self):
            with self._lock:
                if self._state == FUTURE_STATE_FINISHED:
                    exception = self._exception
                    if exception is None:
                        return self._result
                    raise exception
                
                if self._state == FUTURE_STATE_CANCELLED:
                    raise CancelledError
            
            # still pending
            raise InvalidStateError(self, 'result')
    
    @property
    def result(self):
        return self.get_result
    
    
    if __debug__:
        @copy_docs(Future.exception)
        def get_exception(self):
            with self._lock:
                state = self._state
                if state == FUTURE_STATE_FINISHED:
                    self._state = FUTURE_STATE_RETRIEVED
                    return self._exception

                if state == FUTURE_STATE_RETRIEVED:
                    return self._exception

                if state == FUTURE_STATE_CANCELLED:
                    raise CancelledError

            # still pending
            raise InvalidStateError(self, 'exception')

    else:
        @copy_docs(Future.exception)
        def get_exception(self):
            with self._lock:
                if self._state == FUTURE_STATE_FINISHED:
                    return self._exception

                if self._state == FUTURE_STATE_CANCELLED:
                    raise CancelledError

            # still pending
            raise InvalidStateError(self, 'exception')
    
    
    @property
    def exception(self):
        return self.get_exception
    
    
    if __debug__:
        def _done_callback(self, future):
            with self._lock:
                if (self._future is not future):
                    return
                
                state = future._state
                if state == FUTURE_STATE_FINISHED:
                    future._state = FUTURE_STATE_RETRIEVED
                elif state == FUTURE_STATE_RETRIEVED:
                    state = FUTURE_STATE_FINISHED
                
                self._state = state
                self._result = future._result
                self._exception = future._exception
                self._future = None
                self._waiter.set()
    else:
        def _done_callback(self, future):
            with self._lock:
                if (self._future is not future):
                    return
                
                self._state = future._state
                self._result = future._result
                self._exception = future._exception
                self._future = None
                self._waiter.set()
    
    
    set_docs(_done_callback,
        """
        Callback added to the waited future to retrieve it's result.
        
        Parameters
        ----------
        future : ``Future``
            The waited future.
        
        Notes
        -----
        If a different future is given as parameter than the currently waited one, then wont do anything.
        
        If `_debug__` is set as `True`, then this callback also marks the waited future as retrieved.
        """
    )
    
    def _remove_callback(self, future):
        """
        Removes ``._done_callback`` callback from the given `future`'s callbacks.
        
        Parameters
        ----------
        future : ``Future``
            The future, from what's callbacks the own one will be removed.
        """
        callbacks = future._callbacks
        if callbacks:
            for index in range(len(callbacks)):
                callback = callbacks[index]
                if (type(callback) is MethodType) and (callback.__self__ is self):
                    del callbacks[index]
                    break
    
    
    def wait(self, timeout=None, propagate_cancellation=False):
        """
        Waits till the waited future's result or exception is set.
        
        If the future is cancelled, raises ``CancelledError``.
        
        If the future has exception set with `.set_exception`, `.set_exception_if_pending` successfully, then raises
        the given exception.
        
        If the future has result set with `.set_result`, `.set_result_if_pending` successfully, then returns the
        given object.
        
        If the future is not done yet, raises ``InvalidStateError``.
        
        Parameters
        ----------
        timeout : `None`, `float` = `None`, Optional
            Timeout in seconds till the waited future's result should be set. Giving it as `None`, means no time limit.
        propagate_cancellation : `bool` = `False`, Optional
            Whether cancellation should be propagated towards the waited task.
        
        Raises
        ------
        TimeoutError
            If `timeout` is over and the waited future is still pending.
        CancelledError
            The future is cancelled.
        InvalidStateError
            The futures is not done yet.
        TypeError
            The future has non `BaseException` set as exception.
        BaseException
            The future's set exception.
        """
        try:
            result_set = self._waiter.wait(timeout)
        except:
            if propagate_cancellation:
                self.cancel()
            
            raise
        
        if result_set:
            return self.get_result()
        
        if propagate_cancellation:
            self.cancel()
        
        raise TimeoutError
    
    
    def _set_future_result(self, future, result):
        """
        Sets the given `result` as the `future`'s result if applicable.
        
        This method is put on the respective event loop to be called, by ``.set_result`` or by
        ``.set_result_if_pending``, if the future's result can be set.
        
        Parameters
        ----------
        future : ``Future``
            The future to set the result to.
        result : `Any`
            The result to set as the future's.
        """
        try:
            future.set_result(result)
        except (RuntimeError, InvalidStateError): # the future does not support this operation
            pass
        
        with self._lock:
            if self._future is future:
                self._state = FUTURE_STATE_RETRIEVED
                self._future = None
    
    
    @copy_docs(Future.set_result)
    def set_result(self, result):
        with self._lock:
            if self._state != FUTURE_STATE_PENDING:
                raise InvalidStateError(self, 'set_result')
            
            future = self._future
            
            if future is None:
                self._result = result
                self._state = FUTURE_STATE_FINISHED
                self._waiter.set()
                return
        
        loop = self._loop
        loop.call_soon(self.__class__._set_future_result, self, future, result)
        loop.wake_up()
    
    
    @copy_docs(Future.set_result_if_pending)
    def set_result_if_pending(self, result):
        with self._lock:
            if self._state != FUTURE_STATE_PENDING:
                return 0
            
            future = self._future

            if future is None:
                self._result = result
                self._state = FUTURE_STATE_FINISHED
                self._waiter.set()
                return 2

        loop = self._loop
        loop.call_soon(self.__class__._set_future_result, self, future, result)
        loop.wake_up()
        return 1
    
    
    def _set_future_exception(self, future, exception):
        """
        Sets the given `exception` as the `future`'s exception if applicable.
        
        This method is put on the respective event loop to be called, by ``.set_exception`` or by
        ``.set_exception_if_pending``, if the future's result can be set.
        
        Parameters
        ----------
        future : ``Future``
            The future to set the exception to.
        exception : `Any`
            The exception to set as the future's.
        """
        try:
            future.set_exception(exception)
        except (RuntimeError, InvalidStateError): # the future does not supports this operation
            pass
        
        with self._lock:
            if self._future is future:
                self._state = FUTURE_STATE_RETRIEVED
                self._future = None
    
    
    @copy_docs(Future.set_exception)
    def set_exception(self, exception):
        with self._lock:
            if self._state != FUTURE_STATE_PENDING:
                raise InvalidStateError(self, 'set_exception')
            
            if isinstance(exception, type):
                exception = exception()
            
            if isinstance(exception, StopIteration):
                raise TypeError(
                    f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
                )
            
            future = self._future
            if future is None:
                self._exception = exception
                self._state = FUTURE_STATE_FINISHED
                self._waiter.set()
                return
        
        loop = self._loop
        loop.call_soon(self._set_future_exception, future, exception)
        loop.wake_up()
    
    
    @copy_docs(Future.set_exception_if_pending)
    def set_exception_if_pending(self, exception):
        with self._lock:
            if self._state != FUTURE_STATE_PENDING:
                return 0
            
            if isinstance(exception, type):
                exception = exception()
            
            if isinstance(exception, StopIteration):
                raise TypeError(
                    f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
                )
            
            future = self._future
            if future is None:
                self._exception = exception
                self._state = FUTURE_STATE_FINISHED
                self._waiter.set()
                return 2
        
        loop = self._loop
        loop.call_soon(self._set_future_exception, future, exception)
        loop.wake_up()
        return 1
    
    
    if __debug__:
        def __del__(self):
            """
            If the future is pending, but it's result was not set, meanwhile anything waits at it, notifies it.
            
            Also notifies if the future's result was set with ``.set_exception``, or with
            ``.set_exception_if_pending``, and it was not retrieved.
            
            Notes
            -----
            This method is present only if `__debug__` is set as `True`.
            """
            if self._state == FUTURE_STATE_PENDING:
                if self._future is not None:
                    sys.stderr.write(
                        f'{self.__class__.__name__} is not finished, but still pending!\n{self!r}\n'
                    )
                return
            
            if self._state == FUTURE_STATE_FINISHED:
                if (self._exception is not None):
                    write_exception_maybe_async(
                        self._exception,
                        [
                            self.__class__.__name__,
                            ' exception was never retrieved\n',
                            repr(self),
                            '\n',
                        ]
                    )
                return
            
            # no more notify case
        
        def __silence__(self):
            """
            Silences the future's `__del__`, so it will not notify if it would.
            
            Notes
            -----
            This method is present only if `__debug__` is set as `True`.
            """
            self._state = FUTURE_STATE_RETRIEVED
    
    
    @copy_docs(Future.clear)
    def clear(self):
        with self._lock:
            future = self._future
            if future is not None:
                loop = self._loop
                loop.call_soon(self._remove_callback, future)
                loop.wake_up()
                self._future = None
            
            self._waiter.clear()
            self._state = FUTURE_STATE_PENDING
            self._exception = None
            self._result = None



__all__ = ('collect_package_local_variables', 'create_banner', 'create_exit_message')


import sys

from ... import __package__ as PACKAGE_NAME
from ...core import create_event_loop, get_event_loop
from ...utils import HIGHLIGHT_TOKEN_TYPES
from ...utils.trace import _iter_highlight_producer

from .editors.compilation import PYTHON_COMPILE_FLAG_ALLOW_TOP_LEVEL_AWAIT


PACKAGE = __import__(PACKAGE_NAME)


if PYTHON_COMPILE_FLAG_ALLOW_TOP_LEVEL_AWAIT:
    AWAIT_NOTE = 'Use "await" directly.'
else:
    AWAIT_NOTE = '!!! Direct "await" is not available on your python version. Please use python 3.8 or newer !!!'


def get_or_create_event_loop():
    """
    Tries to get the current event loop. If not found creates a new one.
    
    Returns
    -------
    event_loop : ``EventThread``
    """
    try:
        event_loop = get_event_loop()
    except RuntimeError:
        event_loop = create_event_loop(daemon=False)
    
    return event_loop


def _produce_banner(package, logo):
    """
    Produces banner parts with their tokens.
    
    This method is an iterable generator.
    
    Yields
    ------
    part : `str`
        Banner part.
    token_type : `int`
        The part's type.
    """
    if package is None:
        package = PACKAGE
    
    if (logo is None) and (package is PACKAGE):
        logo = (
            f'                     _      _   _\n'
            f'                    | |    | | (_)\n'
            f'  ___  ___ __ _ _ __| | ___| |_ _  ___\n'
            f' / __|/ __/ _` | \'__| |/ _ \ __| |/ _ \\\n'
            f' \__ \ (_| (_| | |  | |  __/ |_| | (_) |\n'
            f' |___/\___\__,_|_|  |_|\___|\__|_|\___/\n'
        )
    
    
    if (logo is not None):
        yield logo, HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_BANNER_LOGO
        
        if (package is not None):
            yield '\n', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE
            
            
            logo_length = max((len(line) for line in logo.splitlines()), default=0)
            package_version = package.__version__
            package_version_length = len(package_version)
            
            adjust = logo_length - package_version_length
            if adjust > 0:
                yield ' ' * adjust, HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE
            
            yield package_version, HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_BANNER_LOGO_VERSION
            yield '\n', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE
            
        yield '\n\n', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE
    
    # At cpython it has a ` \n` in it, but in pypy we have just a `\n`.
    system_version = ' '.join(sys.version.split())
    
    # get package name and upper case it's first character
    package_name = package.__package__
    if package_name and not package_name[0].isupper():
        package_name = package_name[0].upper() + package_name[1:]
    
    console_description = f'{package_name} interactive console {system_version} on {sys.platform}.'
    yield console_description, HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_BANNER_DESCRIPTION
    yield '\n', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE
    
    if PYTHON_COMPILE_FLAG_ALLOW_TOP_LEVEL_AWAIT:
        await_token_type = HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_BANNER_DESCRIPTION_AWAIT
    else:
        await_token_type = HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_BANNER_DESCRIPTION_AWAIT_UNAVAILABLE
    
    yield AWAIT_NOTE, await_token_type
    
    yield '\n', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE
    console_help ='Type "help", "copyright", "credits" or "license" for more information.'
    yield console_help, HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE_BANNER_DESCRIPTION
    yield '\n', HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_CONSOLE


def create_banner(package=None, logo=None, *, highlighter=None):
    """
    helper creating console banner.
    
    Parameters
    ----------
    package : `None`, `ModuleType` = `None`, Optional
        The package, which runs the interactive console.
    
    logo : `None`, `str` = `None`, Optional
        The logo of the package if any.
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    
    Returns
    -------
    banner : `str`
        Console banner.
    """
    return ''.join([*_iter_highlight_producer(_produce_banner(package, logo), highlighter)])


def create_exit_message(package=None):
    """
    helper creating console exit message.
    
    Parameters
    ----------
    package : `None`, `ModuleType` = `None`, Optional
        The package, which runs the interactive console.
    
    Returns
    -------
    exit_message : `str`
        Console exit message.
    """
    if package is None:
        package = PACKAGE
    
    return f'exiting {package.__package__} interactive_console...'


def collect_package_local_variables(package):
    """
    Collects importable variables of scarletio.
    
    Parameters
    ----------
    package : `ModuleType`
        The package to collect variables from.
    
    Returns
    -------
    interactive_console_locals : `dict` of (`str`, `Any`) items
    """
    interactive_console_locals = {package.__package__: package}
    for variable_name in {
        '__name__',
        '__package__',
        '__loader__',
        '__spec__',
        '__builtins__',
        '__file__'
    }:
        interactive_console_locals[variable_name] = getattr(package, variable_name)
    
    for variable_name in getattr(package, '__all__', ()):
        interactive_console_locals[variable_name] = getattr(package, variable_name)
    
    return interactive_console_locals



__all__ = ('DOCS_ENABLED', 'copy_docs',  'has_docs', 'set_docs', )

from functools import partial as partial_func


# Future reference
DOCS_ENABLED = True


def has_docs(target):
    """
    Modifies the target object's docstring if applicable.
    
    Parameter
    ---------
    target : `Any`
        The object to modify the docstring of.
    
    Returns
    -------
    target : `Any`
        The target object.
    """
    if not DOCS_ENABLED:
        target.__doc__ = None
    
    return target

if DOCS_ENABLED and (has_docs.__doc__ is None):
    DOCS_ENABLED = False

if not DOCS_ENABLED:
    has_docs.__doc__ = None


@has_docs
def set_docs(target, docs):
    """
    Sets docstring to the target object.
    
    Parameter
    ---------
    target : `Any`
        The object to set the docstring to.
    docs : `str`
        Docstring to set.
    
    Returns
    -------
    target : `Any`
        The target object.
    """
    if DOCS_ENABLED:
        target.__doc__ = docs
    
    return target


@has_docs
def _do_copy_docs(source, target):
    """
    Does the actual doc-string copy described by ``copy_docs``.
    
    Parameters
    ----------
    source : `Any`
        The source object to copy it's doc-string from.
    target : `Any`
        The target object to copy it's doc-string to.
    
    Returns
    -------
    target : `Any`
        The target object.
    """
    target.__doc__ = source.__doc__
    return target


@has_docs
def copy_docs(source):
    """
    Copies a function's doc-string to an other one.
    
    Parameters
    ----------
    source : `Any`
        The source object to copy it's docs from.
    
    Returns
    -------
    wrapper : ``functools.partial``
        Wrapper which will change the target object's docstring.
    
    Examples
    --------
    
    ```py
    def eat():
        \"\"\"Cakes are great\"\"\"
        pass
    
    @copy_docs(eat)
    def taste():
        pass
    ```
    """
    return partial_func(_do_copy_docs, source)



__all__ = ('SSLBidirectionalTransportLayer',)

import reprlib
from collections import deque
from ssl import SSLError, create_default_context as create_default_ssl_context

from ...utils import copy_docs

from .abstract import AbstractBidirectionalTransportLayerBase
from .extra_info import (
    EXTRA_INFO_NAME_CIPHER, EXTRA_INFO_NAME_COMPRESSION, EXTRA_INFO_NAME_PEER_CERTIFICATION,
    EXTRA_INFO_NAME_SSL_CONTEXT, EXTRA_INFO_NAME_SSL_OBJECT, get_has_extra_info, set_extra_info
)
from .ssl_pipe import SSLPipe
from .transport_layer import TransportLayerBase


class SSLBidirectionalTransportLayer(TransportLayerBase, AbstractBidirectionalTransportLayerBase):
    """
    Asynchronous SSL protocol implementation on top of a `socket`. Uses `MemoryBIO`-s for incoming and
    outgoing data.
    
    Attributes
    ----------
    _extra : `None`, `dict` of (`str`, `Any`) items
        Optional transport information.
    _loop : ``EventThread``
        The event loop to what the transport is bound to.
    _call_connection_made : `bool`
        Whether the the ``.app_protocol``'s `.connection_made` should be called when handshake is completed.
    _closing : `bool`
        Whether the ssl protocol is shut or shutting down.
    _connection_made_waiter : `None`, ``Future``
        A waiter future, what's result is set when connection is made, aka handshake is completed, or if when the
        connection is lost, depending which happens first.
        
        After the future's result or exception is set, the attribute is set as `None`.
    _in_handshake : `bool`
        Whether the ssl transport is in handshaking.
    _protocol : ``AbstractTransportLayerBase``
        Asynchronous protocol implementation.
    _server_host_name : `None`, `str`
        The ssl protocol's server hostname if applicable.
    _server_side : `bool`
        Whether the ssl protocol is server side.
    _session_established : `bool`
        Whether the session is established. Is set after handshake and is set back to `False` when the connection is
        lost.
    _ssl_context : `ssl.SSLContext`
        The connection's ssl type.
    _ssl_pipe : `None`, ``SSLPipe``
        Ssl pipe set meanwhile the protocol is connected to feed the ssl data to.
    _transport : `None`, ``AbstractTransportLayerBase``
        Asynchronous transport implementation.
    _write_backlog : `deque` of `tuple` (`bytes-like`, `int`)
        Ensured data queued up to be written. Each element contains a tuple of the data to write and an offset till
        write from.
    """
    __slots__ = (
        '_call_connection_made', '_closing', '_connection_made_waiter', '_in_handshake', '_protocol',
        '_server_host_name', '_server_side', '_session_established', '_ssl_context', '_ssl_pipe', '_transport',
        '_write_backlog'
    )
    
    def __new__(cls, loop, protocol, ssl_context, connection_made_waiter, server_side, server_host_name,
            call_connection_made):
        """
        Creates a new ``SSLProtocol``.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop of the protocol.
        protocol : ``AbstractTransportLayerBase``
            Asynchronous protocol implementation.
        ssl_context : `None`, `ssl.SSLContext`
            The connection ssl type. If SSl was given as `True` when creating a connection at this point we get it as
            `None`, so we create a default ssl context.
            
            Note, that if the connection is server side, a valid `ssl_context` should be given.
        connection_made_waiter : `None`, ``Future``
            A waiter future, what's result is set when connection is made, aka handshake is completed, or if when the
            connection is lost, depending which happens first.
            
            After the future's result or exception is set, the attribute is set as `None`.
        server_side : `bool`
            Whether the ssl protocol is server side.
        server_host_name : `None`, `str`
            The ssl protocol's server hostname if applicable.
            
            If we are the `server_side`, then this parameter is forced to `None` (wont raise).
        call_connection_made : `bool`
            Whether the the `protocol`'s `.connection_made` should be called when handshake is completed.
        
        Raises
        ------
        ValueError
            If the protocol is server side, but `ssl_context` is not given.
        """
        if ssl_context is None:
            if server_side:
                raise ValueError('Server side SSL needs a valid `SSLContext`.')
            
            ssl_context = create_default_ssl_context()
            if (server_host_name is None) or (not server_host_name):
                ssl_context.check_hostname = False
        
        if server_side:
            server_host_name = None
        
        extra = set_extra_info(None, EXTRA_INFO_NAME_SSL_CONTEXT, ssl_context)
        self = TransportLayerBase.__new__(cls, loop, extra)
        
        self._server_side = server_side
        self._server_host_name = server_host_name
        self._ssl_context = ssl_context
        self._write_backlog = deque()
        self._connection_made_waiter = connection_made_waiter
        self._protocol = protocol
        self._ssl_pipe = None
        self._session_established = False
        self._in_handshake = False
        self._closing = False
        self._transport = None
        self._call_connection_made = call_connection_made
        
        return self
    
    
    def __repr__(self):
        """Returns the ssl protocol transport's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        if self._closing:
            repr_parts.append(' closing')
            field_added = True
        else:
            field_added = False
        
        transport = self._transport
        if (transport is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' transport=')
            repr_parts.append(transport.__class__.__name__)
        
        protocol = self._protocol
        if (protocol is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' protocol=')
            repr_parts.append(protocol.__class__.__name__)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _wake_up_connection_made_waiter(self, exception=None):
        """
        Wakes up the ssl protocol's ``._connection_made_waiter`` if applicable.
        
        Called when connection is made closed or lost, depending, which happens first. If connection is made or closed,
        then the connection made waiter's result is set as `None`, if the connection is lost, then the respective
        exception is thrown into it if any.
        
        Parameters
        ----------
        exception : `None`, ``BaseException`` = `None`, Optional
            Exception to throw into ``._connection_made_waiter`` if any.
        """
        connection_made_waiter = self._connection_made_waiter
        if (connection_made_waiter is not None):
            self._connection_made_waiter = None
            if connection_made_waiter.is_pending():
                if exception is None:
                    connection_made_waiter.set_result(None)
                else:
                    connection_made_waiter.set_exception(exception)
    
    
    @copy_docs(AbstractBidirectionalTransportLayerBase.connection_made)
    def connection_made(self, transport):
        self._transport = transport
        self._ssl_pipe = SSLPipe(self._ssl_context, self._server_side, self._server_host_name)
        
        self._in_handshake = True
        # `(b'', 1)` is a special value in ``._process_write_backlog`` to do the SSL handshake
        self._write_backlog.append((b'', 1))
        self._loop.call_soon(self._process_write_backlog)
    
    
    @copy_docs(AbstractBidirectionalTransportLayerBase.connection_lost)
    def connection_lost(self, exception):
        if self._session_established:
            self._session_established = False
            app_protocol = self._protocol
            self._loop.call_soon(type(app_protocol).connection_lost, app_protocol, exception)
        
        self._transport = None
        self._wake_up_connection_made_waiter(exception)
    
    
    @copy_docs(AbstractBidirectionalTransportLayerBase.pause_writing)
    def pause_writing(self):
        self._protocol.pause_writing()
    
    
    @copy_docs(AbstractBidirectionalTransportLayerBase.resume_writing)
    def resume_writing(self):
        self._protocol.resume_writing()
    
    
    @copy_docs(AbstractBidirectionalTransportLayerBase.data_received)
    def data_received(self, data):
        try:
            ssl_data, application_data = self._ssl_pipe.feed_ssl_data(data)
        except SSLError:
            self.abort()
            return
        
        for chunk in ssl_data:
            self._transport.write(chunk)
        
        for chunk in application_data:
            if chunk:
                self._protocol.data_received(chunk)
                continue
            
            self.close()
            break
    
    
    @copy_docs(AbstractBidirectionalTransportLayerBase.eof_received)
    def eof_received(self):
        try:
            self._wake_up_connection_made_waiter(ConnectionResetError)
            if not self._in_handshake:
                # has no effect whatever it returns when we use ssl
                self._protocol.eof_received()
        finally:
            self._transport.close()
        
        return False
    
    
    @copy_docs(TransportLayerBase.get_extra_info)
    def get_extra_info(self, name, default=None):
        info, present = get_has_extra_info(self._extra, name)
        if not present:
            transport = self._transport
            if transport is not None:
                info = transport.get_extra_info(name, default)
            else:
                info = default
        
        return info
    
    
    def _write_application_data(self, data):
        """
        Writes data to the ``SSLProtocol`` to be sent.
        
        Parameters
        ----------
        data : `bytes-like`
            The data to write.
        """
        self._write_backlog.append((data, 0))
        self._process_write_backlog()
    
    
    def _handshake_completed(self, handshake_exception):
        """
        Called when ssl handshake is completed.
        
        Parameters
        ----------
        handshake_exception : `None`, `BaseException`
            Exception occurred when processing backlog meanwhile handshaking.
        """
        self._in_handshake = False
        ssl_object = self._ssl_pipe._ssl_object
        
        if (handshake_exception is None):
            try:
                peer_certification = ssl_object.getpeercert()
            except BaseException as err:
                handshake_exception = err
            else:
                handshake_exception = None
        
        if (handshake_exception is not None):
            self._transport.close()
            self._wake_up_connection_made_waiter(handshake_exception)
            raise
        
        # Add extra info that becomes available after handshake.
        extra = self._extra
        extra = set_extra_info(extra, EXTRA_INFO_NAME_PEER_CERTIFICATION, peer_certification)
        extra = set_extra_info(extra, EXTRA_INFO_NAME_CIPHER, ssl_object.cipher())
        extra = set_extra_info(extra, EXTRA_INFO_NAME_COMPRESSION, ssl_object.compression())
        extra = set_extra_info(extra, EXTRA_INFO_NAME_SSL_OBJECT, ssl_object)
        self._extra = extra
        
        if self._call_connection_made:
            self._protocol.connection_made(self)
        
        self._wake_up_connection_made_waiter()
        self._session_established = True
        
        # Don's call it immediately, other tasks might be scheduled already.
        self._loop.call_soon(self._process_write_backlog)
    
    
    def _process_write_backlog(self):
        """
        Try to make progress on the write backlog.
        
        Feeds data to ``.ssl_pipe`` till it is full.
        """
        transport = self._transport
        if transport is None:
            return
        
        write_backlog = self._write_backlog
        try:
            for _ in range(len(write_backlog)):
                data, offset = write_backlog[0]
                if data:
                    ssl_data, offset = self._ssl_pipe.feed_application_data(data, offset)
                else:
                    # If data is given as empty bytes means we either have handshake complete or there is no more data
                    # to write
                    if offset:
                        ssl_data = self._ssl_pipe.do_handshake(self._handshake_completed)
                    else:
                        ssl_data = self._ssl_pipe.shutdown(self._finalize)
                    
                    offset = 1
                
                for chunk in ssl_data:
                    transport.write(chunk)
                
                if offset < len(data):
                    write_backlog[0] = (data, offset)
                    # A short write means that a write is blocked on a read, we need to enable reading if it is paused!
                    assert self._ssl_pipe._need_ssl_data
                    if transport._paused:
                        transport.resume_reading()
                    
                    break
                
                # An entire chunk from the backlog was processed. We can delete it and reduce the outstanding buffer
                # size.
                del self._write_backlog[0]
        
        except BaseException as err:
            if self._in_handshake:
                self._handshake_completed(err)
            else:
                self._fatal_error(err, 'Fatal error on SSL transport.')
            
            if not isinstance(err, Exception):
                # BaseException
                raise
    
    
    @copy_docs(TransportLayerBase._fatal_error)
    def _fatal_error(self, exception, message='Fatal error on transport.'):
        TransportLayerBase._fatal_error(self, exception, message)
        
        transport = self._transport
        if (transport is not None):
            transport._force_close(exception)
    
    
    def _finalize(self):
        """
        Closes the ``SSLProtocol``'s transport.
        
        Called after shutdown or abortion.
        """
        transport = self._transport
        if (transport is not None):
            transport.close()
    
    
    def __del__(self):
        """
        Closes the ssl protocol transport if not yet closed.
        """
        self.close()
    
    
    @copy_docs(TransportLayerBase.get_protocol)
    def get_protocol(self):
        return self._protocol
    
    
    @copy_docs(TransportLayerBase.set_protocol)
    def set_protocol(self, protocol):
        self._protocol = protocol
    
    
    @copy_docs(TransportLayerBase.is_closing)
    def is_closing(self):
        return self._closing
    
    
    @copy_docs(TransportLayerBase.close)
    def close(self):
        if not self._closing:
            self._closing = True
            self._write_application_data(b'')
    
    
    @copy_docs(TransportLayerBase.abort)
    def abort(self):
        transport = self._transport
        if (transport is not None):
            try:
                transport.abort()
            finally:
                self._finalize()


    @copy_docs(TransportLayerBase.write)
    def write(self, data):
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(
                f'`data` can be `bytes-like`, got {data.__class__.__name__}; {reprlib.repr(data)}.'
            )
        
        if data:
            self._write_application_data(data)
    
    
    @copy_docs(TransportLayerBase.write_eof)
    def write_eof(self):
        pass
    
    
    @copy_docs(TransportLayerBase.can_write_eof)
    def can_write_eof(self):
        return False
    
    
    @copy_docs(TransportLayerBase.pause_reading)
    def pause_reading(self):
        transport = self._transport
        if transport is None:
            return False
        
        return transport.pause_reading()
    
    
    @copy_docs(TransportLayerBase.resume_reading)
    def resume_reading(self):
        transport = self._transport
        if transport is None:
            return False
            
        transport.resume_reading()
        return True
    
    
    @copy_docs(TransportLayerBase.get_write_buffer_size)
    def get_write_buffer_size(self):
        transport = self._transport
        if transport is None:
            write_buffer_size = 0
        else:
            write_buffer_size = transport.get_write_buffer_size()
        
        return write_buffer_size
    
    
    @copy_docs(TransportLayerBase.get_write_buffer_limits)
    def get_write_buffer_limits(self):
        transport = self._transport
        if transport is None:
            return 16384, 65536
        
        return transport.get_write_buffer_limits()
    
    
    @copy_docs(TransportLayerBase.set_write_buffer_limits)
    def set_write_buffer_limits(self, low=None, high=None):
        transport = self._transport
        if (transport is not None):
            transport.set_write_buffer_limits(high, low)



__all__ = ('AsyncLifoQueue', 'AsyncQueue',)

from collections import deque

from ...utils import copy_docs, copy_func, ignore_frame, to_coroutine

from ..exceptions import CancelledError

from .future import Future


ignore_frame(__spec__.origin, 'get_result_no_wait', 'raise exception',)
ignore_frame(__spec__.origin, '__aexit__', 'raise exception',)

class AsyncQueue:
    """
    An asynchronous FIFO queue.
    
    ``AsyncQueue`` is async iterable, so if you iterate over it inside of an `async for` loop do
    `.set_exception(CancelledError())` to stop it without any specific exception.
    
    Attributes
    ----------
    _exception : `None`, `BaseException`
        The exception set as the queue's result to raise, when the queue gets empty.
    
    _loop : ``EventThread``
        The loop to what the queue is bound to.
    
    _results : `deque`
        The results of the queue, which can be retrieved by ``.result``, ``.result_no_wait``, or by awaiting it.
    
    _set_result_waiters : `None`, `list` of ``Future``
        Result setter waiters for the queue to become empty.
    
    _waiter : `None`, ``Future``
        If the queue is empty and it's result is already waited, then this future is set. It's result is set, by the
        first ``.set_result``, ``.set_exception`` call.
    """
    __slots__ = ('_exception', '_loop', '_results', '_set_result_waiters', '_waiter',)
    
    def __new__(cls, loop, iterable=None, max_length=None, exception=None):
        """
        Creates a new ``AsyncQueue`` with the given parameter.
        
        Parameters
        ----------
        loop : ``EventThread``
            The loop to what the created queue will be bound to.
        iterable : `None`, `iterable` = `None`, Optional
            A preset iterable to extend the queue with.
        max_length : `None`, `int` = `None`, Optional
            The maximal length of the queue.
        exception : `None`, `BaseException` = `None`, Optional
            Exception to raise when the queue is empty.
        
        Raises
        ------
        TypeError
            If `StopIteration` is given as `exception`.
        """
        if (exception is not None):
            if isinstance(exception, type):
                exception = exception()
            
            if isinstance(exception, StopIteration):
                raise TypeError(
                    f'{exception} cannot be raised to a(n) `{cls.__name__}`.'
                )
        
        if iterable is None:
            results = deque(maxlen=max_length)
        else:
            results = deque(iterable, maxlen=max_length)
        
        self = object.__new__(cls)
        self._loop = loop
        self._results = results
        self._waiter = None
        self._exception = exception
        self._set_result_waiters = None
        return self
    
    
    def set_result(self, element):
        """
        Puts the given `element` on the queue. If the queue is empty and it's result is already waited, feeds it to
        ``._waiter`` instead.
        
        Parameters
        ----------
        element : `Any`
            The object to put on the queue.
        
        Returns
        -------
        set_result_state : `int` (`0`, `1`, `2`)
            If the result was set instantly, return `0`. If the result was not set, returns `1`.
        """
        # should we raise InvalidStateError?
        waiter = self._waiter
        if waiter is None:
            results = self._results
            max_length = results.maxlen
            if (max_length is None) or (len(results) < max_length):
                results.append(element)
                set_result = 0
            else:
                set_result = 1
        else:
            self._waiter = None
            waiter.set_result_if_pending(element)
            set_result = 0
        
        return set_result
    
    
    async def set_result_wait(self, element):
        """
        Puts the given `element` on the queue. If the queue is full, blocks till it's elements are exhausted.
        
        This method is an awaitable.
        
        Parameters
        ----------
        element : `Any`
            The object to put on the queue.
        
        Returns
        -------
        set_result_state : `int` (`0`, `1`, `2`)
            If the result was set instantly, return `0`. If the result was not set, returns `1`. If you needed to wait
            for setting the result, returns `2`.
        """
        waiter = self._waiter
        if waiter is None:
            results = self._results
            max_length = results.maxlen
            if (max_length is None) or (len(results) < max_length):
                results.append(element)
                set_result = 0
            else:
                set_result_waiters = self._set_result_waiters
                if (set_result_waiters is None):
                    set_result_waiters = []
                    self._set_result_waiters = set_result_waiters
                
                waiter = Future(self._loop)
                
                set_result_waiters.append(waiter)
                
                try:
                    can_set_result = await waiter
                except:
                    try:
                        set_result_waiters.remove(waiter)
                    except ValueError:
                        pass
                    else:
                        if not set_result_waiters:
                            self._set_result_waiters = None
                    
                    raise
                
                else:
                    if can_set_result:
                        results.append(element)
                        set_result = 2
                    else:
                        set_result = 1
                
        else:
            self._waiter = None
            waiter.set_result_if_pending(element)
            set_result = 0
        
        return set_result
    
    
    def set_exception(self, exception):
        """
        Sets the given `exception` to raise, when it's queue gets empty. If the queue is empty and it's result is
        already waited, feeds it to ``._waiter`` instead.
        
        Parameters
        ----------
        exception : `None`, `BaseException`
            Exception to raise when the queue is empty.
        
        Raises
        ------
        TypeError
            If `StopIteration` is given as `exception`.
        """
        # should we raise InvalidStateError?
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
            )
        
        self._exception = exception
        
        waiter = self._waiter
        if (waiter is not None):
            self._waiter = None
            waiter.set_exception_if_pending(exception)
        
        # cancel all waiters
        set_result_waiters = self._set_result_waiters
        if (set_result_waiters is not None):
            self._set_result_waiters = None
            
            while set_result_waiters:
                set_result_waiters.pop(0).set_result_if_pending(False)
            
    
    def __await__(self):
        """
        Waits till the next element of the queue is set. If the queue has elements set, yields the next of them, or if
        the queue has exception set, raises it.
        
        This method is a generator. Should be used with `await` expression.
        
        Returns
        -------
        result : `Any`
            The next element on the queue.
        
        Raises
        ------
        BaseException
            Exception set to the queue, to raise when it is empty.
        """
        results = self._results
        if results:
            return results.popleft()
        
        exception = self._exception
        if exception is not None:
            raise exception
        
        waiter = self._waiter
        if waiter is None:
            waiter = Future(self._loop)
            self._waiter = waiter
        
        try:
            return (yield from waiter)
        finally:
            self._poll_from_set_result_waiters()
    
    get_result = to_coroutine(copy_func(__await__))
    
    @property
    def result(self):
        return self.get_result
    
    
    def _poll_from_set_result_waiters(self):
        """
        Polls one future from set result waiters.
        """
        set_result_waiters = self._set_result_waiters
        if (set_result_waiters is not None):
            while True:
                set_result = set_result_waiters.pop(0).set_result_if_pending(True)
                
                if (not set_result_waiters):
                    self._set_result_waiters = None
                    break
                
                if set_result == 1:
                    break
    
    
    def get_result_no_wait(self):
        """
        Returns the queue's next element if applicable.
        Waits till the next element of the queue is set. If the queue has elements set, yields the next of them, or if
        the queue has exception set, raises it.
        
        Returns
        -------
        result : `Any`
            The next element on the queue.
        
        Raises
        ------
        IndexError
            The queue is empty.
        BaseException
            Exception set to the queue, to raise when it is empty.
        """
        results = self._results
        if results:
            result = results.popleft()
            self._poll_from_set_result_waiters()
            return result
        
        exception = self._exception
        if exception is None:
            raise IndexError('The queue is empty')
        
        raise exception
    
    
    @property
    def result_no_wait(self):
        return self.get_result_no_wait
    
    
    def __repr__(self):
        """Returns the async queue's representation."""
        repr_parts = [
            self.__class__.__name__,
            '([',
        ]
        
        results = self._results
        limit = len(results)
        if limit:
            index = 0
            while True:
                element = results[index]
                repr_parts.append(repr(element))
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
            
        repr_parts.append(']')
        
        max_length = results.maxlen
        if (max_length is not None):
            repr_parts.append(', max_length=')
            repr_parts.append(repr(max_length))
        
        exception = self._exception
        if (exception is not None):
            repr_parts.append(', exception=')
            repr_parts.append(str(exception))
        
        repr_parts.append(')')
        return ''.join(repr_parts)
    
    __str__ = __repr__
    
    def __aiter__(self):
        """
        Async iterating over an ``AsyncQueue``, returns itself
        
        Returns
        -------
        self : ``AsyncQueue``
        """
        return self
    
    
    async def __anext__(self):
        """
        Waits till the next element of the queue is set. If the queue has elements set, yields the next of them, or if
        the queue has exception set, raises it.
        
        If the queue has ``CancelledError`` set as ``._exception``, then raises ``StopAsyncIteration`` to stop the queue
        instead.
        
        This method is a coroutine.
        
        Returns
        -------
        result : `Any`
            The next element on the queue.
        
        Raises
        ------
        StopAsyncIteration
            If the queue was cancelled with ``CancelledError``.
        BaseException
            Exception set to the queue, to raise when it is empty.
        """
        results = self._results
        if results:
            result = results.popleft()
            self._poll_from_set_result_waiters()
            return result
        
        exception = self._exception
        if exception is not None:
            if type(exception) is CancelledError:
                raise StopAsyncIteration from CancelledError
            
            raise exception
        
        waiter = self._waiter
        if waiter is None:
            waiter = Future(self._loop)
            self._waiter = waiter
        
        try:
            return (await waiter)
        except CancelledError as err:
            raise StopAsyncIteration from err
    
    # deque operations
    
    @property
    def max_length(self):
        """
        Returns the queue's max length.
        
        Returns
        -------
        max_length: `int`
        """
        return self._results.maxlen
    
    
    def clear(self):
        """
        Clears the queue's results.
        """
        self._results.clear()
    
    
    def copy(self):
        """
        Copies the queue.
        
        Returns
        -------
        new : ``AsyncQueue``
        """
        new = object.__new__(type(self))
        
        new._loop = self._loop
        new._results = self._results.copy()
        new._waiter = None
        new._exception = self._exception
        new._set_result_waiters = None
        
        return new
    
    
    def reverse(self):
        """
        Reverses the queue's actual results.
        """
        self._results.reverse()
    
    
    def __len__(self):
        """
        Returns the queue's actual length.
        """
        return len(self._results)
    
    
    if __debug__:
        def __del__(self):
            """
            If the queue has ``_waiter`` set, silences it.
            
            Notes
            -----
            This function is only present, when `__debug__` is set as `True`.
            """
            waiter = self._waiter
            if waiter is not None:
                waiter.__silence__()


class AsyncLifoQueue(AsyncQueue):
    """
    An asynchronous LIFO queue.
    
    ``AsyncLifoQueue`` is async iterable, so if you iterate over it inside of an `async for` loop do
    `.set_exception(CancelledError())` to stop it without any specific exception.
    
    Attributes
    ----------
    _exception : `None`, `BaseException`
        The exception set as the queue's result to raise, when the queue gets empty.
    _loop : ``EventThread``
        The loop to what the queue is bound to.
    _results : `deque`
        The results of the queue, which can be retrieved by ``.result``, ``.result_no_wait``, or by awaiting it.
    _waiter : `None`, ``Future``
        If the queue is empty and it's result is already waited, then this future is set. It's result is set, by the
        first ``.set_result``, ``.set_exception`` call.
    """
    __slots__ = ()
    
    @copy_docs(AsyncQueue.__await__)
    def __await__(self):
        results = self._results
        if results:
            return results.pop()
        
        exception = self._exception
        if exception is not None:
            raise exception
        
        waiter = self._waiter
        if waiter is None:
            waiter = Future(self._loop)
            self._waiter = waiter
        
        return (yield from waiter)
    
    @copy_docs(AsyncQueue.result_no_wait)
    def get_result_no_wait(self):
        results = self._results
        if results:
            return results.pop()
        
        exception = self._exception
        if exception is None:
            raise IndexError('The queue is empty')
        
        raise exception
    
    
    @property
    def result_no_wait(self):
        return self.get_result_no_wait
    
    
    @copy_docs(AsyncQueue.__anext__)
    async def __anext__(self):
        results = self._results
        if results:
            return results.pop()
        
        exception = self._exception
        if exception is not None:
            if type(exception) is CancelledError:
                raise StopAsyncIteration from CancelledError
            
            raise exception
        
        waiter = self._waiter
        if waiter is None:
            waiter = Future(self._loop)
            self._waiter = waiter
        
        try:
            return (await waiter)
        except CancelledError as err:
            raise StopAsyncIteration from err



import pathlib, re
from ast import literal_eval
from setuptools import setup


HERE = pathlib.Path(__file__).parent

# Lookup version
version_search_pattern = re.compile('^__version__[ ]*=[ ]*((?:\'[^\']+\')|(?:\"[^\"]+\"))[ ]*$', re.M)
parsed = version_search_pattern.search((HERE / 'scarletio' / '__init__.py').read_text())
if parsed is None:
    raise RuntimeError('No version found in `__init__.py`.')

version = literal_eval(parsed.group(1))

# Lookup readme
README = (HERE / 'README.md').read_text('utf-8')

setup(
    name = 'scarletio',
    version = version,
    packages = [
        'scarletio',
        'scarletio.core',
        'scarletio.ext',
        'scarletio.ext.asyncio',
        'scarletio.core.event_loop',
        'scarletio.core.subprocess',
        'scarletio.core.top_level',
        'scarletio.core.protocols_and_transports',
        'scarletio.core.traps',
        'scarletio.http_client',
        'scarletio.tools',
        'scarletio.tools.asynchronous_interactive_console',
        'scarletio.tools.asynchronous_interactive_console.editors',
        'scarletio.utils',
        'scarletio.utils.highlight',
        'scarletio.web_common',
        'scarletio.websocket',
    ],
    url = 'https://github.com/HuyaneMatsu/scarletio',
    license = 'DBAD',
    author = 'HuyaneMatsu',
    author_email = 're.ism.tm@gmail.com',
    description = 'Asynchronous blackmagic & Witchcraft',
    long_description = README,
    long_description_content_type = 'text/markdown',
    classifiers = [
        'Development Status :: 4 - Beta',
        
        'Intended Audience :: Developers',
        
        'Operating System :: OS Independent',
        
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    include_package_data = True,
    package_data = {
    },
    python_requires = '>=3.6',
    install_requires = [
        'chardet>=2.0',
    ],
    extras_require = {
        'cpythonspeedups': [
            'cchardet>=2.0',
        ],
    },
)
